# Copyright 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#            http://www.apache.org/licenses/LICENSE-2.0
#
#            Unless required by applicable law or agreed to in writing, software
#            distributed under the License is distributed on an "AS IS" BASIS,
#            WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#            See the License for the specific language governing permissions and
#            limitations under the License.
"""Control flow graph analysis.

Given a Python AST we construct a doubly linked control flow graph whose nodes
contain the AST of the statements. We can then perform forward analysis on this
CFG.

"""
from collections import Counter
import inspect
import functools
import operator
from copy import copy
from queue import Queue

import ast as gast

LITERALS = (gast.Num, gast.Str, gast.Bytes, gast.Ellipsis, gast.NameConstant)

CONTROL_FLOW = (gast.For, gast.AsyncFor, gast.While, gast.If, gast.Try, gast.Break, gast.Continue,
                gast.With, gast.AsyncWith)

COMPOUND_STATEMENTS = (gast.FunctionDef, gast.ClassDef, gast.For, gast.While, gast.If, gast.With,
                       gast.Try, gast.AsyncFunctionDef, gast.AsyncFor, gast.AsyncWith)

SIMPLE_STATEMENTS = (gast.Return, gast.Delete, gast.Assign, gast.AugAssign, gast.Raise, gast.Assert,
                     gast.Import, gast.ImportFrom, gast.Global, gast.Nonlocal, gast.Expr, gast.Pass,
                     gast.Break, gast.Continue)

STATEMENTS = COMPOUND_STATEMENTS + SIMPLE_STATEMENTS

BLOCKS = (
    (gast.Module, 'body'),
    (gast.FunctionDef, 'body'),
    (gast.AsyncFunctionDef, 'body'),
    (gast.For, 'body'),
    (gast.For, 'orelse'),
    (gast.AsyncFor, 'body'),
    (gast.AsyncFor, 'orelse'),
    (gast.While, 'body'),
    (gast.While, 'orelse'),
    (gast.If, 'body'),
    (gast.If, 'orelse'),
)


def get_name(node):
    """Get the name of a variable.
    Args:
        node: A `Name`, `Subscript` or `Attribute` node.
    Returns:
        The name of the variable e.g. `'x'` for `x`, `x.i` and `x[i]`.
    """
    if isinstance(node, gast.Name):
        return node.id
    elif isinstance(node, (gast.Subscript, gast.Attribute)):
        return get_name(node.value)
    else:
        raise TypeError


def _get_target(node):
    if isinstance(node, (gast.Name, gast.Subscript, gast.Attribute)):
        return set([get_name(node)])
    elif isinstance(node, (gast.Tuple, gast.List)):
        return set.union(*(_get_target(target) for target in node.elts))
    else:
        raise ValueError


def get_updated(node):
    """Return the variable names created or mutated by this statement.
    This function considers assign statements, augmented assign statements, and
    the targets of for loops, as well as function arguments.
    For example, `x[0] = 2` will return `x`, `x, y = 3, 4` will return `x` and
    `y`, `for i in range(x)` will return `i`, etc.
    Args:
        node: An AST node
    Returns:
        A set of variable names (strings) of all the variables created or mutated.
    """
    if isinstance(node, gast.Assign):
        return set.union(*(_get_target(target) for target in node.targets))
    elif isinstance(node, (gast.For, gast.AugAssign)):
        return _get_target(node.target)
    elif isinstance(node, gast.AsyncWith):
        return set.union(*(_get_target(target.optional_vars) for target in node.items))
    elif isinstance(node, gast.AsyncFor):
        return _get_target(node.target)
    elif isinstance(node, gast.With):
        breakpoint()
    elif isinstance(node, gast.arguments):
        targets = set(arg.arg for arg in node.args + node.kwonlyargs)
        if node.vararg:
            targets.add(node.vararg.arg)
        if node.kwarg:
            targets.add(node.kwarg.arg)
        return targets
    else:
        return set()


class Loop(gast.AST):
    pass


class Node(object):
    """A node in the CFG."""
    __slots__ = ['next', 'value', 'prev', 'loop']

    def __init__(self, value, loop=False):
        self.loop = loop
        self.next = []
        self.prev = []
        self.value = value


class CFG(gast.NodeVisitor):
    """Construct a control flow graph.

    Each statement is represented as a node. For control flow statements such
    as conditionals and loops the conditional itself is a node which either
    branches or cycles, respectively.

    Attributes:
        entry: The entry node, which contains the `gast.arguments` node of the
                function definition.
        exit: The exit node. This node is special because it has no value (i.e. no
                corresponding AST node). This is because Python functions can have
                multiple return statements.
    """
    def __init__(self):
        # The current leaves of the CFG
        self.head = []
        # A stack of continue statements
        self.continue_ = []
        # A stack of break nodes
        self.break_ = []

    @staticmethod
    def backlink(node):
        """Given a CFG with outgoing links, create incoming links."""
        seen = set()
        to_see = [node]
        while to_see:
            node = to_see.pop()
            seen.add(node)
            for succ in node.next:
                succ.prev.append(node)
                if succ not in seen:
                    to_see.append(succ)

    def set_head(self, node):
        """Link this node to the current leaves."""
        for head in self.head:
            head.next.append(node)
        self.head[:] = []
        self.head.append(node)

    @classmethod
    def build_cfg(cls, node):
        """Build a CFG for a function.

        Args:
            node: A function definition the body of which to analyze.

        Returns:
            A CFG object.

        Raises:
            TypeError: If the input is not a function definition.
        """
        if not isinstance(node, (gast.FunctionDef, gast.AsyncFunctionDef, gast.Lambda)):
            breakpoint()
            raise TypeError('input must be a function definition')
        cfg = cls()
        cfg.entry = Node(node.args)
        cfg.head = [cfg.entry]
        cfg.visit_statements(node.body)
        cfg.exit = Node(None)
        cfg.set_head(cfg.exit)
        cfg.backlink(cfg.entry)
        return cfg

    def visit_statements(self, nodes):
        for node in nodes:
            if isinstance(node, CONTROL_FLOW):
                self.visit(node)
            else:
                expr = Node(node)
                self.set_head(expr)

    def generic_visit(self, node):
        raise ValueError('unknown control flow')

    def visit_If(self, node):
        # The current head will hold the conditional
        test = Node(node.test)
        self.set_head(test)
        # Handle the body
        self.visit_statements(node.body)
        body_exit = self.head[:]
        self.head[:] = []
        self.head.append(test)
        # Handle the orelse
        self.visit_statements(node.orelse)
        self.head.extend(body_exit)

    def visit_AsyncWith(self, node):
        # The current head will hold the conditional
        test = Node(node)
        self.set_head(test)
        # Handle the body
        self.visit_statements(node.body)

    visit_With = visit_AsyncWith

    def Loop(self, body, orelse, test):
        test = Node(test)
        self.set_head(test)
        loop = Node(Loop())
        self.set_head(loop)
        # Start a new level of nesting
        self.break_.append([])
        self.continue_.append([])
        # Handle the body
        self.visit_statements(body)
        self.head.extend(self.continue_.pop())

        test_exit = Node(copy(test.value))
        self.set_head(test_exit)
        self.set_head(loop)
        # test.next[0].loop = True
        # Handle the orelse
        self.visit_statements(orelse)
        # The break statements and the test go to the next node
        self.head = [test, test_exit]
        self.head.extend(self.break_.pop())

    def visit_While(self, node):
        self.Loop(node.body, node.orelse, node.test)

    def visit_AsyncFor(self, node):
        self.Loop(node.body, node.orelse, node.iter)

    def visit_For(self, node):
        self.Loop(node.body, node.orelse, node.iter)

    def visit_Break(self, node):
        self.break_[-1].extend(self.head)
        self.head[:] = []

    def visit_Continue(self, node):
        self.continue_[-1].extend(self.head)
        self.head[:] = []

    def visit_Try(self, node):
        self.visit_statements(node.body)
        body = self.head
        handlers = []
        for handler in node.handlers:
            self.head = body[:]
            self.visit_statements(handler.body)
            handlers.extend(self.head)
        self.head = body
        self.visit_statements(node.orelse)
        self.head = handlers + self.head
        self.visit_statements(node.finalbody)


def _loops_on_self_rec(start_node, cur_node, reaching):
    if cur_node not in reaching:
        return False

    names = [v[0] for v in reaching[start_node]['gen']]
    for name, in_node in reaching[cur_node]['in']:
        if in_node is cur_node:
            continue

        if name in names:
            if in_node is start_node:
                return True

            if _loops_on_self_rec(start_node, in_node, reaching):
                return True

    return False


def loops_on_self(node, reaching):
    return _loops_on_self_rec(node, node, reaching)


class Forward(object):
    """Forward analysis on CFG.

  Args:
    label: A name for this analysis e.g. 'active' for activity analysis. The
        AST nodes in the CFG will be given annotations 'name_in', 'name_out',
        'name_gen' and 'name_kill' which contain the incoming values, outgoing
        values, values generated by the statement, and values deleted by the
        statement respectively.
    gen: A function which takes the CFG node as well as a set of incoming
        values. It must return a set of newly generated values by the statement
        as well as a set of deleted (killed) values.
    op: Either the AND or OR operator. If the AND operator is used it turns
        into forward must analysis (i.e. a value will only be carried forward
        if it appears on all incoming paths). The OR operator means that
        forward may analysis is done (i.e. the union of incoming values will be
        taken).
  """
    def __init__(self, label, gen, op=operator.or_):
        self.gen = gen
        self.op = op
        self.out_label = label + '_out'
        self.in_label = label + '_in'
        self.gen_label = label + '_gen'
        self.kill_label = label + '_kill'
        self.reaching = {}
        self.registers = set()

    def visit(self, node):
        if node not in self.reaching:
            self.reaching[node] = {}

        reaching = self.reaching[node]

        if node.value:
            if 'out' in reaching:
                in_before = reaching['in']
            else:
                in_before = None

            preds = [
                self.reaching[pred]['out'] for pred in node.prev
                if 'out' in self.reaching.get(pred, {})
            ]
            if preds:
                incoming = functools.reduce(self.op, preds[1:], preds[0])
            else:
                incoming = frozenset()

            reaching['in'] = incoming
            gen, kill = self.gen(node, incoming)
            reaching['gen'] = gen
            reaching['kill'] = kill
            reaching['out'] = (incoming - kill) | gen

            if reaching['in'] != in_before:
                for succ in node.next:
                    self.visit(succ)
        else:
            preds = [
                self.reaching[pred]['out'] for pred in node.prev
                if 'out' in self.reaching.get(pred, {})
            ]
            self.exit = functools.reduce(self.op, preds[1:], preds[0])


class InferRegisters:
    def __init__(self, reaching):
        self.reaching = reaching
        self.visited = set()
        self.registers = set()

    def visit(self, node):
        if node in self.visited:
            return

        if isinstance(node.value, Loop):
            # Check if any of the variables are conditionaly changed (i.e.
            # under some if statement) in the loop. This means that their value
            # needs to be registered.
            loop_end = [p for p in node.prev if p not in self.visited]
            loop_start = [p for p in node.prev if p in self.visited]

            all_in_end = set()
            for n in loop_end:
                all_in_end |= self.reaching[n]['out']

            all_in_start = set()
            for n in loop_start:
                all_in_start |= self.reaching[n]['out']

            changed = set(name for name, n in (all_in_end - all_in_start))
            reached_from_before = set(name for name, _ in (all_in_end & all_in_start)
                                      if name in changed)

            self.registers |= reached_from_before

        if not isinstance(node.value, Loop) and any(p not in self.visited for p in node.prev):
            return

        # print(f'Visit: {node.value}')
        self.visit_ast(node.value, self.reaching[node])
        self.visited.add(node)

        for succ in node.next:
            self.visit(succ)

    def visit_ast(self, node, reaching):
        for base_class in inspect.getmro(node.__class__):
            if hasattr(self, base_class.__name__):
                getattr(self, base_class.__name__)(node, reaching)
                break
        else:
            self.generic_visit(node, reaching)

    def generic_visit(self, node, reaching):
        # print(f'Generic: {node}')
        pass

    def expr(self, node, reaching):
        if all(d[1] in self.visited for d in reaching.get('in', [])):
            return

        variables = []
        for succ in gast.walk(node):
            if not isinstance(succ, gast.Name):
                continue

            if not (isinstance(succ.ctx, gast.Load) or isinstance(node, gast.AugAssign)):
                continue

            variables.append(succ.id)


        for name, n in reaching['in']:
            if (n in self.visited) or (name not in variables) or (name in self.registers):
                continue

            self.registers.add(name)

    def Expr(self, node: gast.Expr, reaching):
        self.visit_ast(node.value, reaching)

    def Return(self, node: gast.Yield, reaching):
        self.expr(node.value, reaching)

    def Yield(self, node: gast.Yield, reaching):
        self.expr(node.value, reaching)

    def AugAssign(self, node: gast.AugAssign, reaching):
        self.expr(node, reaching)

    def Assign(self, node: gast.Assign, reaching):
        self.expr(node, reaching)

    AnnAssign = Assign

    def AsyncWith(self, node: gast.AsyncWith, reaching):
        # TODO: Fix me!
        pass


def dfs(node, G, visited, labels):
    if node in visited:
        return

    visited.add(node)
    labels[id(node)] = type(node.value).__name__

    for n in node.next:
        G.add_edge(id(node), id(n))

        dfs(n, G, visited, labels)


def draw_cfg(cfg):
    import networkx as nx
    import matplotlib.pyplot as plt

    G = nx.DiGraph()
    visited = set()
    labels = {}

    dfs(cfg.entry, G, visited, labels)

    pos = nx.planar_layout(G)
    nx.draw(G, pos, font_size=16, with_labels=False)

    # for p in pos:  # raise text positions
    #     pos[p][1] += 0.07

    nx.draw_networkx_labels(G, pos, labels)
    plt.show()


def forward(node, analysis):
    """Perform a given analysis on all functions within an AST."""
    if not isinstance(analysis, Forward):
        raise TypeError('not a valid forward analysis object')

    cfg_obj = CFG.build_cfg(node)
    # draw_cfg(cfg_obj)

    analysis.visit(cfg_obj.entry)

    v = InferRegisters(analysis.reaching)
    v.visit(cfg_obj.entry)

    # print(f'Registers: {v.registers}')

    return node, cfg_obj.entry, analysis.reaching, v.registers


def reaching(node):
    return forward(node, ReachingDefinitions())


class ReachingDefinitions(Forward):
    """Perform reaching definition analysis.

    Each statement is annotated with a set of (variable, definition) pairs.

    """
    def __init__(self):
        def definition(node, incoming):
            definitions = get_updated(node.value)
            gen = frozenset((id_, node) for id_ in definitions)
            kill = frozenset(def_ for def_ in incoming if def_[0] in definitions)
            return gen, kill

        super(ReachingDefinitions, self).__init__('definitions', definition)



# from pygears.hls.hls_expressions import EXTENDABLE_OPERATORS, BinOpExpr
from pygears.hls import ir
from pygears.hdl.sv.sv_expression import SVExpressionVisitor, sieve_slices, get_slice_func
from pygears.typing import Array, Number, typeof, is_type, Tuple, Union, Integral, Queue, Integer

RESIZE_FUNC_TEMPLATE = """
function {signed} [{res_size}:0] {name};
    input {signed} [{op_size}:0] tmp;
    begin
        {name} = tmp;
    end
endfunction
"""


def get_resize_func(res_dtype, op_dtype):
    res_size = res_dtype.width
    op_size = op_dtype.width
    signed = 'signed' if getattr(op_dtype, 'signed', False) else ''

    name = f'ext_{op_size}_to_{res_size}_{signed}'
    val = RESIZE_FUNC_TEMPLATE.format(res_size=res_size - 1,
                                      op_size=op_size - 1,
                                      signed=signed,
                                      name=name)
    return name, val


def update_extras(all_extras, addition):
    if addition:
        for key, value in addition.items():
            if key:
                all_extras[key] = value


def cast(res_dtype, op_dtype, op_value):
    truncate_func = None
    truncate_impl = None

    if isinstance(op_value, int):
        return str(op_value), {truncate_func: truncate_impl}

    res = op_value

    if op_dtype.width != res_dtype.width:
        truncate_func, truncate_impl = get_resize_func(res_dtype, op_dtype)
        res = f'{truncate_func}({res})'

    res_signed = getattr(res_dtype, 'signed', False)
    op_signed = getattr(op_dtype, 'signed', False)

    if res_signed != op_signed:
        sign = '$signed' if res_signed else '$unsigned'
        res = f'{sign}({res})'

    return res, {truncate_func: truncate_impl}


class VExpressionVisitor(SVExpressionVisitor):
    def __init__(self, aux_funcs=None):
        super(VExpressionVisitor, self).__init__(aux_funcs=aux_funcs)
        self.separator = '.'
        self.expr = vexpr
        self.extras = {}

    def cast_svexpr(self, svexpr, expr_dtype, cast_dtype):
        expr_signed = getattr(expr_dtype, 'signed', False)
        res_signed = getattr(cast_dtype, 'signed', False)

        expr_width = expr_dtype.width
        cast_width = cast_dtype.width

        if cast_width == 0:
            return None

        if res_signed != expr_signed:
            if res_signed:
                svexpr = f"$signed({{1'b0, {svexpr}}})"
                expr_width += 1
            else:
                svexpr = f"$unsigned({svexpr})"

        if cast_width != expr_width:
            truncate_func, truncate_impl = get_resize_func(cast_dtype, expr_dtype)
            self.aux_funcs[truncate_func] = truncate_impl
            svexpr = f'{truncate_func}({svexpr})'

        if res_signed:
            svexpr = f"$signed({svexpr})"

        return svexpr


def vexpr(expr, aux_funcs=None):
    v_visit = VExpressionVisitor(aux_funcs)
    return v_visit.visit(expr)



from pygears import gear, GearDone
from pygears.core.gear import GearTypeNotSpecified
from pygears.typing import Int, Tuple, Uint, bitw, Unit
from pygears.typing.base import TypingMeta


def get_int_type(val):
    if val == 0:
        return Uint[1]
    elif val > 0:
        return Uint[bitw(val)]
    else:
        return Int[bitw(val)]


def get_literal_type(val):
    if isinstance(type(val), TypingMeta):
        return type(val)
    elif isinstance(val, int):
        return get_int_type(val)
    elif isinstance(val, tuple):
        dtypes = [get_literal_type(v) for v in val]
        return Tuple[tuple(dtypes)]
    else:
        raise GearTypeNotSpecified(
            f"Value {val} not supported for const module")


@gear(hdl={'impl': 'sustain'})
async def const(*, val, tout=b'get_literal_type(val)') -> b'tout':
    yield tout(val)


@gear
async def once(*, val, tout=b'get_literal_type(val)') -> b'tout':
    yield tout(val)

    while True:
        raise GearDone


@gear
async def fix(din, *, val, tout=b'get_literal_type(val)') -> b'tout':
    async with din:
        yield tout(val)

ping = fix(val=Unit())


@gear(hdl={'impl': 'empty'})
async def void(*, dtype) -> b'dtype':
    raise GearDone



from pygears.typing import Queue, typeof


def din_data_cat(intfs, order=None):

    if order is None:
        order = range(len(intfs))

    data = []
    for o in order:
        intf = intfs[o]

        if intf['modport'] == 'producer':
            continue

        if issubclass(intf['type'], Queue):
            if intf['type'][0].width > 0:
                data.append(f'{intf["name"]}_s.data')
        else:
            if intf['type'].width > 0:
                data.append(f'{intf["name"]}_s')

    return f'{{ {", ".join(reversed(data))} }}'


def din_data_cat_value(data):
    dout = []
    for d in data:
        if isinstance(d, Queue):
            dout.append(d.data)
        else:
            dout.append(d)

    return tuple(dout)



from pygears import alternative, gear
from pygears.typing import Queue, Union, Uint, Tuple, Bool, Unit, Any, Maybe
from pygears.lib.fmaps.queue import queuemap
from .ccat import ccat


def filt_type(din, lvl, sel):
    return Queue[(din[0].types)[sel], lvl]


@gear
async def filt(din: Tuple[{'data': Union, 'sel': Uint}]) -> b'din["data"]':
    '''Filters the ``data`` field of :class:`Union` type, by passing it forward
    only if it carries the ``data`` :class:`Union` subtype whose index is equal
    to the value supplied to the ``sel`` field. Index of a :class:`Union`
    subtype is equal to the position of the subtype within the :class:`Union`
    definition.

    '''

    async with din as (data, sel):
        if data.ctrl == sel:
            yield data


@alternative(filt)
@gear
def filt_unit(din: Union[Any, Unit]) -> b'din.types[0]':
    return din | filt(fixsel=0)


@alternative(filt)
@gear
def filt_maybe(din: Union[Unit, Any]) -> b'din.types[1]':
    return din | filt(fixsel=1)


@alternative(filt)
@gear
def filt_fix_sel(din: Union, *, fixsel) -> b'din.types[fixsel]':
    return (ccat(din, din.dtype.ctrl(fixsel)) | filt)[0] \
        >> din.dtype.types[fixsel]


@alternative(filt)
@gear
def qfilt_f(din: Queue, *, f):
    @gear
    def maybe_out(din, *, f):
        return ccat(din, din | f) | Union

    return din | queuemap(f=maybe_out(f=f)) | qfilt_union(fixsel=1)


@alternative(filt)
@gear(enablement=b'not typeof(din, Queue)')
def filt_f(din, *, f):
    @gear
    def maybe_out(din, *, f):
        return ccat(din, f(din)) | Union

    return din | maybe_out(f=f) | filt(fixsel=1)


@alternative(filt)
@gear
async def qfilt_union(din: Queue[Union, 'lvl'], *, fixsel=0,
                      filt_lvl=1) -> b'filt_type(din, lvl, fixsel)':

    data_reg: din.dtype.data.data = din.dtype.data.data(0)
    eot_reg: Uint[din.dtype.lvl] = Uint[din.dtype.lvl](0)
    empty_reg: Bool = Bool(True)
    curr_data: din.dtype.data.data
    field_sel: Bool

    while True:
        async with din as d:
            curr_data = d.data.data
            field_sel = (d.data.ctrl == fixsel)

            if all(d.eot[:filt_lvl]):
                if field_sel:
                    if not empty_reg:
                        yield (data_reg, eot_reg)
                        yield (curr_data, d.eot)
                elif not empty_reg:
                    yield (data_reg, d.eot)
                empty_reg = True

            elif field_sel:
                if not empty_reg:
                    yield (data_reg, eot_reg)

                # register
                data_reg = curr_data
                eot_reg = d.eot
                empty_reg = False



from dataclasses import dataclass
from pygears.typing.math import ceil_chunk, ceil_pow2


@dataclass
class Port:
    name: str
    direction: str
    width: int

AXI_MASTER = {
    'common': [
        Port('aclk', 'out', 0),
        Port('aresetn', 'out', 0),
    ],
    'awaddr': [
        Port('awaddr', 'out', -1),
        Port('awid', 'out', -1),
        Port('awprot', 'out', 3),
        Port('awvalid', 'out', 0),
        Port('awready', 'in', 0),
        Port('awsize', 'out', 3),
        Port('awburst', 'out', 2),
        Port('awcache', 'out', 4),
        Port('awlen', 'out', 8),
    ],
    'awaddr_qos': [
        Port('awlock', 'out', 1),
        Port('awqos', 'out', 4),
        Port('awregion', 'out', 4),
    ],
    'wdata': [
        Port('wdata', 'out', -1),
        Port('wstrb', 'out', -1),
        Port('wvalid', 'out', 0),
        Port('wready', 'in', 0),
        Port('wlast', 'out', 0),
    ],
    'bresp': [
        Port('bresp', 'in', 2),
        Port('bvalid', 'in', 0),
        Port('bready', 'out', 0),
    ],
    'araddr': [
        Port('araddr', 'out', -1),
        Port('arid', 'out', -1),
        Port('arprot', 'out', 3),
        Port('arvalid', 'out', 0),
        Port('arready', 'in', 0),
        Port('arsize', 'out', 3),
        Port('arburst', 'out', 2),
        Port('arcache', 'out', 4),
        Port('arlen', 'out', 8),
    ],
    'araddr_qos': [
        Port('arlock', 'out', 1),
        Port('arqos', 'out', 4),
        Port('arregion', 'out', 4),
    ],
    'rdata': [
        Port('rdata', 'in', -1),
        Port('rid', 'in', -1),
        Port('rresp', 'in', 2),
        Port('rvalid', 'in', 0),
        Port('rready', 'out', 0),
        Port('rlast', 'in', 0),
    ],
}


AXI_SLAVE = {
    'common': [
        Port('aclk', 'in', 0),
        Port('aresetn', 'in', 0),
    ],
    'awaddr': [
        Port('awaddr', 'in', -1),
        Port('awid', 'in', -1),
        Port('awprot', 'in', 3),
        Port('awvalid', 'in', 0),
        Port('awready', 'out', 0),
        Port('awsize', 'in', 3),
        Port('awburst', 'in', 2),
        Port('awcache', 'in', 4),
        Port('awlen', 'in', 8),
    ],
    'awaddr_qos': [
        Port('awlock', 'in', 1),
        Port('awqos', 'in', 4),
        Port('awregion', 'in', 4),
    ],
    'wdata': [
        Port('wdata', 'in', -1),
        Port('wstrb', 'in', -1),
        Port('wvalid', 'in', 0),
        Port('wready', 'out', 0),
        Port('wlast', 'in', 0),
    ],
    'bresp': [
        Port('bresp', 'out', 2),
        Port('bvalid', 'out', 0),
        Port('bready', 'in', 0),
        Port('bid', 'out', -1),
    ],
    'araddr': [
        Port('araddr', 'in', -1),
        Port('arid', 'in', -1),
        Port('arprot', 'in', 3),
        Port('arvalid', 'in', 0),
        Port('arready', 'out', 0),
        Port('arsize', 'in', 3),
        Port('arburst', 'in', 2),
        Port('arcache', 'in', 4),
        Port('arlen', 'in', 8),
    ],
    'araddr_qos': [
        Port('arlock', 'in', 1),
        Port('arqos', 'in', 4),
        Port('arregion', 'in', 4),
    ],
    'rdata': [
        Port('rdata', 'out', -1),
        Port('rid', 'out', -1),
        Port('rresp', 'out', 2),
        Port('rvalid', 'out', 0),
        Port('rready', 'in', 0),
        Port('rlast', 'out', 0),
    ],
}

AXIL_SLAVE = {
    'common': [
        Port('aclk', 'in', 0),
        Port('aresetn', 'in', 0),
    ],
    'awaddr': [
        Port('awaddr', 'in', -1),
        Port('awprot', 'in', 3),
        Port('awvalid', 'in', 0),
        Port('awready', 'out', 0),
    ],
    'wdata': [
        Port('wdata', 'in', -1),
        Port('wstrb', 'in', -1),
        Port('wvalid', 'in', 0),
        Port('wready', 'out', 0),
    ],
    'bresp': [
        Port('bresp', 'out', 2),
        Port('bvalid', 'out', 0),
        Port('bready', 'in', 0),
    ],
    'araddr': [
        Port('araddr', 'in', -1),
        Port('arprot', 'in', 3),
        Port('arvalid', 'in', 0),
        Port('arready', 'out', 0),
    ],
    'rdata': [
        Port('rdata', 'out', -1),
        Port('rresp', 'out', 2),
        Port('rvalid', 'out', 0),
        Port('rready', 'in', 0),
    ],
}

AXIS_SLAVE = {
    'common': [
        Port('aclk', 'in', 0),
        Port('aresetn', 'in', 0),
    ],
    'tdata': [
        Port('tdata', 'in', -1),
        Port('tvalid', 'in', 0),
        Port('tready', 'out', 0),
        Port('tlast', 'in', 0),
    ]
}

AXIS_MASTER = {
    'common': [
        Port('aclk', 'in', 0),
        Port('aresetn', 'in', 0),
    ],
    'tdata': [
        Port('tdata', 'out', -1),
        Port('tvalid', 'out', 0),
        Port('tready', 'in', 0),
        Port('tlast', 'out', 0),
    ]
}


def subport_def(subintf, prefix, **kwds):
    ret = []
    for p in subintf:
        if kwds.get(p.name, True) is None:
            continue

        width = p.width
        if p.width == -1:
            if p.name in kwds:
                width = kwds[p.name]
            elif 'dflt' in kwds:
                width = kwds['dflt']
            else:
                breakpoint()
                raise Exception(f'Port "{p.name}" wasn\'t supplied a parameter')

            if p.name in ['wdata', 'rdata'] and isinstance(width, int):
                width = ceil_chunk(ceil_pow2(int(width)), 32)
            elif p.name == 'tdata' and isinstance(width, int):
                width = ceil_chunk(ceil_pow2(int(width)), 8)

        direction = 'input' if p.direction == 'in' else 'output'
        name = f'{prefix}_{p.name}'

        if width == 0:
            ret.append(f'{direction} {name}')
        elif isinstance(width, str):
            ret.append(f'{direction} [{width}:0] {name}')
        else:
            ret.append(f'{direction} [{width-1}:0] {name}')

    return ret


def port_def(intf, prefix, **kwds):
    ret = []
    for subintf in intf:
        if subintf in kwds:
            subkwds = {}
            if isinstance(kwds[subintf], dict):
                subkwds = kwds[subintf]
            elif isinstance(kwds[subintf], bool):
                if not kwds[subintf]:
                    continue
            elif isinstance(kwds[subintf], int):
                subkwds = {'dflt': kwds[subintf]}

            ret.extend(subport_def(intf[subintf], prefix, **subkwds))

    return ret


def port_map(intf, prefix_src, prefix_dest, uppercase=False, **kwds):
    ret = []
    for subintf in intf:
        if not kwds.get(subintf, False):
            continue

        cfg = kwds[subintf]

        for p in intf[subintf]:
            if isinstance(cfg, dict) and not cfg.get(p.name, True):
                continue

            dest = f'{prefix_dest}_{p.name}'

            if uppercase:
                dest = dest.upper()

            ret.append(f'.{dest}({prefix_src}_{p.name})')

    return ret



from pygears import gear, alternative, module
from pygears.typing import Union, Tuple, Uint, Any, Unit, Maybe, typeof, TypeMatchError
from pygears.lib.shred import shred
from pygears.lib.const import fix
from pygears.lib.ccat import ccat
from pygears.lib.fmaps.union import unionmap
from pygears.lib.mux import field_mux, mux
from pygears.lib.filt import filt


@gear(enablement=b'len(din) >= 1')
def mux_by(ctrl, *din, fmux=mux):
    return fmux(ctrl, *din) | union_collapse


@gear(enablement=b'len(din) == 2')
def union_sync(*din, ctrl, outsync=True) -> b'din':
    pass


@gear
def union_sync_with(din, sync_in, *, ctrl, balance=None):
    if balance:
        sync_in = sync_in | balance

    sync_in_sync, din_sync = union_sync(sync_in, din, ctrl=ctrl)
    sync_in_sync | shred

    return din_sync


@gear
def case(cond, din, *, f, fcat=ccat, tout=None, mapping=None, **kwds):
    try:
        len(f)
    except TypeError:
        f = (None, f)

    if mapping is None:
        in_num_types = len(f)
    else:
        in_num_types = max(mapping.keys()) + 1

    return fcat(din, cond) \
        | Union[(din.dtype, ) * in_num_types] \
        | unionmap(f=f, mapping=mapping, **kwds) \
        | union_collapse(t=tout)


@gear
def ucase(din: Union, *, f, fcat=ccat, tout=None, fmux=mux, mapping=None):
    dout = din \
        | unionmap(f=f, fmux=fmux, mapping=mapping, use_dflt=False)

    if dout.dtype.types.count(dout.dtype.types[0]) != len(dout.dtype.types):
        raise TypeMatchError(
            f'output types of all input types need be the same, but got "{dout.dtype.types}')

    return dout | union_collapse(t=tout)


@gear
def when(cond, din, *, f, fe, fcat=ccat, tout=None, **kwds):
    return din | case(cond, f=(fe, f), fcat=fcat, tout=tout, **kwds)


@alternative(when)
@gear
def when_single(cond, din, *, f):
    return ccat(din, cond) | Union | filt(fixsel=1) | f


@gear
async def valve(cond, din) -> b'din':
    async with cond as c:
        if c:
            async with din as d:
                yield d


@alternative(when)
@gear
async def when_pass(cond, din, *, halt=False) -> b'din':
    async with cond as c:
        if not halt or c:
            async with din as d:
                if c:
                    yield d


def all_same(din):
    return din.types.count(din.types[0]) == len(din.types)


@gear
def maybe_when(cond, din, *, f, fcat=ccat):
    return fcat(din, cond) \
        | Union \
        | unionmap(f=(fix(val=Unit()), f)) \
        | Maybe


def dflt_map(dtypes):
    return {i: i for i in range(len(dtypes))}


@gear
def select(cond: Uint, *din, mapping=b'dflt_map(din)'):

    dtypes = [d.dtype for d in din]
    if dtypes.count(dtypes[0]) != len(dtypes):
        raise TypeError(
            f'Expected all inputs to "{module().name}" to be same type, but got: "{dtypes}"')

    return mux(cond, *din, mapping=mapping) | union_collapse


@gear
def field_sel(din: Tuple[{'ctrl': Uint, 'data': Any}], *, mapping=b'dflt_map(din["data"])'):

    if typeof(din.dtype['data'], Tuple):
        dtypes = list(din.dtype['data'])
        if dtypes.count(dtypes[0]) != len(dtypes):
            raise TypeError(
                f'Expected all inputs to "{module().name}" to be same type, but got: "{dtypes}"')

    return field_mux(din, mapping=mapping) | union_collapse


@gear(enablement=b'all_same(din) or t')
def union_collapse(din: Union, *, t=None):
    if t is None:
        t = din.dtype.types[0]

    return din[0] >> t



import functools
import typing
import os
import pygears
from pygears import reg, Intf
from ...base_resolver import ResolverBase, ResolverTypeError
from pygears.util.fileio import find_in_dirs, save_file
from pygears.conf import inject, Inject
from pygears.hdl.sv.v.accessors import rewrite


def get_port_config(modport, type_, name):
    return {
        'modport': modport,
        'name': name,
        'size': 1,
        'type': type_,
        'width': type_.width,
        'local_type': type_
    }


class HDLTemplateResolver(ResolverBase):
    @inject
    def __init__(self, node):
        self.node = node

        if self.impl_path is None:
            raise ResolverTypeError

    @property
    def hdl_path_list(self):
        return reg[f'{self.lang}gen/include']

    @property
    def impl_basename(self):
        fn = self.node_def_name
        if 'hdl' in self.node.meta_kwds:
            if 'impl' in self.node.meta_kwds['hdl']:
                fn = self.node.meta_kwds['hdl']['impl']

        if not os.path.splitext(fn)[-1]:
            fn = f'{fn}.{self.lang}t'

        return fn

    @property
    def impl_path(self):
        return find_in_dirs(self.impl_basename, self.hdl_path_list)

    @property
    def files(self):
        files = [self.file_basename]
        if 'hdl' in self.node.meta_kwds:
            if 'files' in self.node.meta_kwds['hdl']:
                for fn in self.node.meta_kwds['hdl']['files']:
                    if not os.path.splitext(fn)[-1]:
                        fn = f'{fn}.{self.lang}'

                    files.append(fn)

        return files

    @property
    def module_name(self):
        return self.hier_path_name

    @property
    def params(self):
        return {}

    @property
    def file_basename(self):
        return f'{self.module_name}.{self.lang}'

    def module_context(self, template_env):
        context = {
            'pygears': pygears,
            'module_name': self.module_name,
            'intfs': template_env.port_intfs(self.node),
            # 'sigs': [s.name for s in self.node.meta_kwds['signals']],
            'sigs': self.node.meta_kwds['signals'],
            'params': self.node.params,
            'inst': [],
            'generics': []
        }

        if 'attrib' in self.cfg:
            attrib = self.cfg.get('attrib', None)
            if isinstance(attrib, str):
                attrib = [attrib]

            context['comment'] = {'attrib': attrib}

        for port in context['intfs']:
            context[f'_{port["name"]}'] = port
            context[f'_{port["name"]}_t'] = port['type']

        return context

    def generate(self, template_env, outdir):
        ctx = self.module_context(template_env)
        module = template_env.render_local(self.impl_path, self.impl_basename,
                                           self.module_context(template_env))

        if template_env.lang == 'v':
            index = {}
            for intf in ctx['intfs']:
                from pygears.hls import ir
                direction = ir.IntfType.iin if intf['modport'] == 'consumer' else ir.IntfType.iout

                index[intf['name']] = ir.IntfType[intf['type'], direction]
                index[f'{intf["name"]}_s'] = intf['type']

            module = rewrite(module, index)

        save_file(self.file_basename, outdir, module)



import copy
import inspect
import itertools
import operator
from functools import reduce, wraps


def safe_nested_set(dictionary, value, *keys):
    '''sets empty dict if there is no key'''
    try:
        # check if path already exists
        reduce(operator.getitem, keys[:-1], dictionary)
    except KeyError:
        for i, key in enumerate(keys[:-1]):
            # set empty dict for missing keys
            try:
                reduce(operator.getitem, keys[:i + 1], dictionary)
            except KeyError:
                nested_set(dictionary, {}, *keys[:i + 1])

    nested_set(dictionary, value, *keys)


def nested_get(dictionary, *keys):
    return reduce(operator.getitem, keys, dictionary)


def nested_set(dictionary, value, *keys):
    nested_get(dictionary, *keys[:-1])[keys[-1]] = value


def dict_generator(indict, pre=None):
    pre = pre[:] if pre else []
    if isinstance(indict, dict):
        for key, value in indict.items():
            if isinstance(value, dict):
                for d in dict_generator(value, pre + [key]):
                    yield d
            elif isinstance(value, list) or isinstance(value, tuple):
                for v in value:
                    for d in dict_generator(v, pre + [key]):
                        yield d
            else:
                yield pre + [key, value]
    else:
        try:
            # check if iterable
            iter(indict)
            yield indict
        except TypeError:
            yield [indict]


def intercept_arguments(func, cb_named=None, cb_kwds=None, cb_pos=None):
    if inspect.isgeneratorfunction(func):
        from .log import conf_log
        conf_log().warning('Intercepting arguments for generator function '
                           'could result in callbacks beeing called to soon')

    sig = inspect.signature(func)
    # default values in func definition
    dflt_args = {
        k: v.default
        for k, v in sig.parameters.items()
        if v.default is not inspect.Parameter.empty
    }
    # *args in func definition
    positional = {
        k: None
        for k, v in sig.parameters.items()
        if v.kind is inspect.Parameter.VAR_POSITIONAL
    }
    # **kwargs in func definition
    keyword = {
        k: None
        for k, v in sig.parameters.items() if v.kind in
        [inspect.Parameter.VAR_KEYWORD, inspect.Parameter.KEYWORD_ONLY]
    }
    # named parameters in func definition
    named = {
        k: None
        for k, v in sig.parameters.items()
        if (k not in positional) and (k not in keyword)
    }

    @wraps(func)
    def wrapper(*args, **kwargs):
        pos_args = []
        kw_args = {}
        named_args = copy.deepcopy(named)

        # get all passed arguments
        passed_args = list(
            itertools.zip_longest(named,
                                  args,
                                  fillvalue=inspect.Parameter.empty))

        # get all named and keyword arguments
        for k, v in passed_args:
            if inspect.Parameter.empty not in [k, v]:
                kwargs[k] = v

        # find passed positional arguments
        pos_args = [
            val for name, val in passed_args if name is inspect.Parameter.empty
        ]

        # use default values if none are passed
        for k, v in dflt_args.items():
            if (k not in kwargs) or (kwargs[k] is inspect.Parameter.empty):
                kwargs[k] = v

        # split named and keyword arguments
        for k, v in kwargs.items():
            if k in named_args:
                named_args[k] = v
            else:
                kw_args[k] = v

        # callbacks
        if cb_named:
            cb_named(named_args)
        if cb_kwds:
            cb_kwds(kw_args)
        if cb_pos:
            cb_pos(pos_args)

        return func(*named_args.values(), *pos_args, **kw_args)

    return wrapper



import ast
import sys
import types
import inspect
from functools import partial
from . import Context, HLSSyntaxError, node_visitor, ir, visit_ast, visit_block
from .arith import resolve_arith_func
from .call import resolve_func, const_func_args
from .stmt import output_port_shadow_check
from pygears.typing import cast, Integer, typeof


@node_visitor(ast.Expr)
def expr(node, ctx: Context):
    return visit_ast(node.value, ctx)


@node_visitor(ast.IfExp)
def parse_ifexp(node, ctx: Context):
    test = visit_ast(node.test, ctx)

    if isinstance(test, ir.ResExpr):
        if test.val:
            return visit_ast(node.body, ctx)
        else:
            return visit_ast(node.orelse, ctx)

    body = visit_ast(node.body, ctx)
    orelse = visit_ast(node.orelse, ctx)

    return ir.ConditionalExpr(operands=(body, orelse), cond=test)


# TODO: Signal use of reserved names. Also use of output port names (Or mangle them?)
@node_visitor(ast.Name)
def _(node: ast.Name, ctx: Context):
    if isinstance(node.ctx, ast.Load):
        if (node.id in ctx.alias_map and isinstance(ctx.alias_map[node.id], ir.ResExpr)
                and node.id not in ctx.registers):
            return ctx.alias_map[node.id]

        if node.id not in ctx.scope:
            if node.id in ctx.local_namespace:
                return ir.ResExpr(ctx.local_namespace[node.id])

            builtins = ctx.local_namespace['__builtins__']
            if not isinstance(builtins, dict):
                builtins = builtins.__dict__

            if node.id in builtins:
                return ir.ResExpr(builtins[node.id])

            raise SyntaxError(f"Name '{node.id}' not found")

        return ctx.ref(node.id, ctx='load')

    elif isinstance(node.ctx, ast.Store):
        if node.id in ctx.scope:
            return ctx.ref(node.id, ctx='store')

        return ir.Name(node.id, None, 'store')


@node_visitor(ast.UnaryOp)
def _(node: ast.UnaryOp, ctx: Context):
    operand = visit_ast(node.operand, ctx)

    if type(node.op) in [ast.Not]:
        return ir.UnaryOpExpr(operand, type(node.op))

    # TODO: This WAS needed, since: ir.ResExpr(Uint[8]).dtype is None. REMOVE
    dtype = type(operand.val) if isinstance(operand, ir.ResExpr) else operand.dtype
    # f = getattr(operand.dtype, METHOD_OP_MAP[type(op)])

    f = getattr(dtype, METHOD_OP_MAP[type(node.op)])

    ret = resolve_func(f, (operand, ), {}, ctx)

    if ret != ir.ResExpr(NotImplemented):
        return ret

    raise SyntaxError(f"Operator '{ir.OPMAP[type(node.op)]}' not supported for operand of type "
                      f"{operand.dtype!r}")


@node_visitor(ast.Tuple)
def _(node, ctx: Context):
    return ir.ConcatExpr([visit_ast(item, ctx) for item in node.elts])


@node_visitor(ast.Subscript)
def _(node, ctx: Context):
    expr_ctx = 'load' if isinstance(node.ctx, ast.Load) else 'store'
    return ir.SubscriptExpr(visit_ast(node.value, ctx), visit_ast(node.slice, ctx), expr_ctx)


METHOD_OP_MAP = {
    ast.Add: '__add__',
    ast.And: '__and__',
    ast.BitAnd: '__and__',
    ast.BitOr: '__or__',
    ast.BitXor: '__xor__',
    ast.Div: '__truediv__',
    ast.Eq: '__eq__',
    ast.Gt: '__gt__',
    ast.GtE: '__ge__',
    ast.FloorDiv: '__floordiv__',
    ast.Invert: '__invert__',
    ast.Lt: '__lt__',
    ast.LtE: '__le__',
    ast.LShift: '__lshift__',
    ast.MatMult: '__matmul__',
    ast.Mult: '__mul__',
    ast.Mod: '__mod__',
    ast.NotEq: '__ne__',
    ast.Not: '__not__',
    ast.Or: '__or__',
    ast.RShift: '__rshift__',
    ast.Sub: '__sub__',
    ast.UAdd: '__pos__',
    ast.USub: '__neg__',
}

R_METHOD_OP_MAP = {
    ast.Add: '__radd__',
    ast.BitAnd: '__rand__',
    ast.BitOr: '__ror__',
    ast.BitXor: '__rxor__',
    ast.Div: '__rtruediv__',
    ast.Eq: '__eq__',
    ast.Gt: '__le__',
    ast.GtE: '__lt__',
    ast.FloorDiv: '__rfloordiv__',
    ast.In: '__contains__',
    ast.Lt: '__ge__',
    ast.LtE: '__gt__',
    ast.MatMult: '__rmatmul__',
    ast.Mult: '__rmul__',
    ast.Mod: '__rmod__',
    ast.NotEq: '__ne__',
    ast.Sub: '__rsub__'
}


def visit_bin_expr(op, operands, ctx: Context):
    op1 = visit_ast(operands[0], ctx)
    op2 = visit_ast(operands[1], ctx)

    # TODO: This WAS needed, since: ir.ResExpr(Uint[8]).dtype is None. REMOVE
    dtype = type(op1.val) if isinstance(op1, ir.ResExpr) else op1.dtype
    # f = getattr(op1.dtype, METHOD_OP_MAP[type(op)])

    if type(op) in METHOD_OP_MAP:
        f = getattr(dtype, METHOD_OP_MAP[type(op)])

        ret = resolve_func(f, (op1, op2), {}, ctx)

        if ret != ir.ResExpr(NotImplemented):
            return ret

    # TODO: This WAS needed, since: ir.ResExpr(Uint[8]).dtype is None. REMOVE
    dtype = type(op2.val) if isinstance(op2, ir.ResExpr) else op2.dtype
    # f = getattr(op2.dtype, R_METHOD_OP_MAP[type(op)])

    f = getattr(dtype, R_METHOD_OP_MAP[type(op)])

    ret = resolve_func(f, (op2, op1), {}, ctx)

    if ret != ir.ResExpr(NotImplemented):
        return ret

    raise SyntaxError(f"Operator '{ir.OPMAP[type(op)]}' not supported for operands of types "
                      f"{op1.dtype!r} and {op2.dtype!r}")


@node_visitor(ast.Compare)
def _(node, ctx: Context):
    left = node.left
    exprs = []
    for op, c in zip(node.ops, node.comparators):
        exprs.append(visit_bin_expr(op, (left, c), ctx))
        left = c

    if len(exprs) == 1:
        return exprs[0]
    else:
        return ir.BinOpExpr(exprs, ir.opc.And)


@node_visitor(ast.BoolOp)
def _(node, ctx: Context):
    ops = [visit_ast(opi, ctx) for opi in node.values]

    base = ir.BinOpExpr((ops[0], ops[1]), type(node.op))

    for op in ops[2:]:
        base = ir.BinOpExpr((base, op), type(node.op))

    return base


@node_visitor(ast.BinOp)
def _(node, ctx: Context):
    return visit_bin_expr(node.op, (node.left, node.right), ctx)


def get_dict_attr(obj, attr):
    for cls in (obj, ) + inspect.getmro(obj.__class__):
        if attr in cls.__dict__:
            if cls.__name__ == "GenericMeta":
                raise AttributeError
                # raise Exception(f'Cannot parse method {attr}')

            return cls.__dict__[attr]

    raise AttributeError


@node_visitor(ast.JoinedStr)
def _(node, ctx: Context):
    # TODO: Implement this properly
    return ir.ResExpr('')


def attr_shortcuts(cls, attr):
    if typeof(cls, ir.IntfType) and attr == 'dtype':
        return ir.ResExpr(cls.dtype)


@node_visitor(ast.Attribute)
def _(node, ctx: Context):
    value = visit_ast(node.value, ctx)
    if isinstance(value, ir.ResExpr):
        return ir.ResExpr(getattr(value.val, node.attr))

    if hasattr(value.dtype, node.attr):
        res = attr_shortcuts(value.dtype, node.attr)
        if res is not None:
            return res

        try:
            cls_attr = get_dict_attr(value.dtype, node.attr)

            # Maybe we have a class method, try to get function
            cls_attr = getattr(cls_attr, 'func', cls_attr)
        except AttributeError:
            # TODO: When is this necessary
            cls_attr = getattr(value.dtype, node.attr)

        # try:
        #     cls_attr = object.__getattribute__(value.dtype, node.attr).func
        # except Exception:
        #     cls_attr = getattr(value.dtype, node.attr)

        if isinstance(cls_attr, property):
            return resolve_func(cls_attr.fget, (value, ), {}, ctx)

        # TODO: What does this exactly do?
        if not inspect.isclass(cls_attr):
            if callable(cls_attr):
                return ir.ResExpr(partial(cls_attr, value))

        # If value.attr was a class method and we instantly got the result
        return ir.ResExpr(cls_attr)

    return ir.AttrExpr(value, node.attr)


if sys.version_info[1] < 8:

    @node_visitor(ast.Str)
    def _(node, ctx: Context):
        return ir.ResExpr(node.s)

    @node_visitor(ast.NameConstant)
    def _(node, ctx: Context):
        return ir.ResExpr(node.value)

    @node_visitor(ast.Num)
    def num(node, ctx: Context):
        return ir.ResExpr(cast(node.n, Integer))

else:

    @node_visitor(ast.Constant)
    def _(node, ctx: Context):
        if isinstance(node.value, (int, float)):
            return ir.ResExpr(cast(node.n, Integer))
        else:
            return ir.ResExpr(node.value)


@node_visitor(ast.Index)
def _(node, ctx: Context):
    return visit_ast(node.value, ctx)


@node_visitor(ast.Slice)
def _(node, ctx: Context):
    return ir.SliceExpr(visit_ast(node.lower, ctx), visit_ast(node.upper, ctx),
                        visit_ast(node.step, ctx))


def py_eval_expr(node, ctx: Context):
    gear_locals = {n: v.val for n, v in ctx.scope.items() if isinstance(v, ir.ResExpr)}

    return eval(
        compile(ast.Expression(ast.fix_missing_locations(node)), filename="<ast>", mode="eval"),
        gear_locals, ctx.local_namespace)


@node_visitor(ast.DictComp)
def _(node, ctx: Context):
    return ir.ResExpr(py_eval_expr(node, ctx))


@node_visitor(ast.Dict)
def _(node, ctx: Context):
    return ir.ResExpr(py_eval_expr(node, ctx))


# TODO: Generalize for multiple loops withing list comprehension
@node_visitor(ast.ListComp)
def _(node, ctx: Context):
    if len(node.generators) != 1:
        raise SyntaxError(f'Only list comprehensions with a single for loop supported')

    iterator = visit_ast(node.generators[0].iter, ctx)
    target = visit_ast(node.generators[0].target, ctx)

    output_port_shadow_check(target.name, ctx)

    outter_var = None
    outter_namespace = None
    if target.name in ctx.scope:
        outter_var = ctx.scope[target.name]
        del ctx.scope[target.name]

    if target.name in ctx.local_namespace:
        outter_namespace = ctx.local_namespace[target.name]
        del ctx.local_namespace[target.name]

    stmts = []
    if isinstance(iterator, ir.ResExpr):
        vals = iterator.val
    elif isinstance(iterator, ir.CallExpr):
        if not const_func_args(iterator.args.values(), iterator.kwds):
            raise Exception(f'Generator needs to have constant arguments')

        args = tuple(node.val for node in iterator.args.values())
        vals = list(iterator.func(*args))
    else:
        vals = [ir.SubscriptExpr(iterator, ir.ResExpr(i)) for i in range(len(iterator.dtype))]

    for v in vals:
        ctx.local_namespace[target.name] = v
        res = visit_ast(node.elt, ctx)
        if isinstance(res, list):
            if len(res) > 1:
                raise SyntaxError(f'Complex expressions not supported for list comprehensions')

            res = res[0]

        stmts.append(res)

    del ctx.local_namespace[target.name]

    if outter_var:
        ctx.scope[target.name] = outter_var

    if outter_namespace:
        ctx.local_namespace[target.name] = outter_namespace

    return ir.ConcatExpr(stmts)


@node_visitor(ast.List)
def _(node: ast.List, ctx: Context):
    return ir.ResExpr(ir.ConcatExpr([visit_ast(e, ctx) for e in node.elts]))



from pygears import reg
from pygears.hdl.modinst import HDLModuleInst
from .sv_keywords import sv_keywords
from collections import OrderedDict
from pygears.conf import inject, Inject
from pygears.hdl import hdlmod, mod_lang, rename_ambiguous


class SVModuleInst(HDLModuleInst):
    @property
    def inst_name(self):
        inst_name = super().inst_name

        if inst_name in sv_keywords:
            return f'{inst_name}_i'
        else:
            return inst_name

    def get_impl_wrap(self, template_env):
        intfs = template_env.port_intfs(self.node)

        port_map = {}
        for i in intfs:
            name = i['name']
            port_map[f'{name}_data'] = f'{name}.data'
            port_map[f'{name}_valid'] = f'{name}.valid'
            port_map[f'{name}_ready'] = f'{name}.ready'

        context = {
            'wrap_module_name': self.module_name,
            'module_name': self.module_basename,
            'inst_name': f'{self.module_basename}_i',
            'intfs': intfs,
            'sigs': self.node.meta_kwds['signals'],
            'param_map': self.params,
            'port_map': port_map
        }

        return template_env.render_local(__file__, "impl_wrap.j2", context)

    def get_wrap_portmap(self, parent_lang):
        sig_map = {}
        for s in self.node.meta_kwds['signals']:
            sig_map[s.name] = s.name

        port_map = {}
        for p in self.node.in_ports + self.node.out_ports:
            name = p.basename
            if self.lang == 'sv':
                port_map[name] = name
            elif parent_lang == 'sv':
                sig_map[f'{name}_valid'] = f'{name}.valid'
                sig_map[f'{name}_ready'] = f'{name}.ready'
                sig_map[f'{name}_data'] = f'{name}.data'
            else:
                port_map[name] = name

        return port_map, sig_map

    def get_wrap(self, parent_lang):
        template_env = reg[f'{parent_lang}gen/templenv']

        port_map, sigmap = self.get_wrap_portmap(parent_lang)

        context = {
            'rst_name': 'rst',
            'wrap_module_name': self.wrap_module_name,
            'module_name': self.module_name,
            'inst_name': self.inst_name,
            'param_map': self.params,
            'port_map': port_map,
            'intfs': template_env.port_intfs(self.node),
            'sigs': self.node.meta_kwds['signals'],
            'sig_map': sigmap
        }

        return template_env.render(template_env.basedir, f"{self.lang}_wrap.j2", context)

    def get_synth_wrap(self, template_env):
        context = {
            'wrap_module_name': f'wrap_{self.module_name}',
            'module_name': self.module_name,
            'inst_name': self.inst_name,
            'intfs': template_env.port_intfs(self.node),
            'sigs': self.node.meta_kwds['signals'],
            'param_map': self.resolver.params
        }

        return template_env.render('.', "module_synth_wrap.j2", context)

    def get_out_port_map_intf_name(self, port, lang):
        basename = hdlmod(port.consumer).basename
        if lang == 'sv':
            return basename
        else:
            return basename, None

    def get_in_port_map_intf_name(self, port, lang):
        intf = port.producer
        hdlgen_intf = hdlmod(intf)

        if len(intf.consumers) == 1:
            if lang == 'sv':
                return hdlgen_intf.outname
            else:
                return hdlgen_intf.outname, None
        else:
            i = intf.consumers.index(port)
            if lang == 'sv':
                return f'{hdlgen_intf.outname}[{i}]'
            else:
                return (hdlgen_intf.outname, i)

    def get_inst(self, template_env, port_map=None):
        parent_lang = mod_lang(self.node.parent)
        module_name = self.wrap_module_name

        if parent_lang == self.lang:
            params = self.params
        else:
            template_env = reg[f'{parent_lang}gen/templenv']
            params = {}

        if not port_map:
            in_port_map = [(port.basename,
                            self.get_in_port_map_intf_name(port, parent_lang))
                           for port in self.node.in_ports]

            out_port_map = [(port.basename,
                             self.get_out_port_map_intf_name(port, parent_lang))
                            for port in self.node.out_ports]
            port_map = OrderedDict(in_port_map + out_port_map)

        sigmap = {}
        for s in self.node.meta_kwds['signals']:
            sigmap[s.name] = self.node.params['sigmap'].get(s.name, s.name)

        context = {
            'rst_name': 'rst',
            'module_name': rename_ambiguous(module_name, self.lang),
            'inst_name': self.inst_name,
            'param_map': params,
            'port_map': port_map,
            'sig_map': sigmap
        }

        return template_env.snippets.module_inst(**context)



import inspect
from functools import wraps


def doublewrap(f):
    '''
    a decorator decorator, allowing the decorator to be used as:
    @decorator(with, arguments, and=kwds)
    or
    @decorator
    '''
    @wraps(f)
    def new_dec(*args, **kwds):
        if len(args) == 1 and len(kwds) == 0 and callable(args[0]):
            # actual decorated function
            return f(args[0])
        else:
            # decorator arguments
            return lambda realf: f(realf, *args, **kwds)

    return new_dec


def perpetum(func, *args, **kwds):
    while True:
        yield func(*args, **kwds)


def is_standard_func(func):
    return not (inspect.iscoroutinefunction(func) or inspect.isgeneratorfunction(func)
                or is_async_gen(func))


def is_async_gen(func):
    return bool(func.__code__.co_flags & inspect.CO_ASYNC_GENERATOR)


def get_function_context_dict(func):
    if func.__closure__ is None:
        return func.__globals__

    context = {}
    context.update(func.__globals__)

    for name, cell in zip(func.__code__.co_freevars, func.__closure__):
        try:
            context[name] = cell.cell_contents
        except ValueError:
            # Cell is empty
            pass

    return context



import unittest
import subprocess
import inspect
import os
import re
import shutil
from functools import partial, wraps

import jinja2
import pytest

from pygears import clear, find, reg
from pygears.sim import sim
from pygears.sim.modules.sim_socket import SimSocket
from pygears.sim.modules.verilator import SimVerilated
from pygears.hdl import register_hdl_paths
from pygears.hdl import hdlgen, synth
from pygears.hdl import yosys

re_trailing_space_rem = re.compile(r"\s+$", re.MULTILINE)
re_multispace_rem = re.compile(r"\s+", re.MULTILINE)
re_multi_comment_rem = re.compile(r"/\*.*?\*/", re.DOTALL)
re_comment_rem = re.compile(r"//.*$", re.MULTILINE)

rem_pipe = {
    re_multi_comment_rem: '',
    re_comment_rem: '',
    re_multispace_rem: ' ',
    re_trailing_space_rem: ''
}


def remove_unecessary(s):

    for r, repl in rem_pipe.items():
        s = r.sub(repl, s)

    return s.strip()


def equal_on_nonspace(str1, str2):
    # print(remove_unecessary(str1))
    # print('-----------------------')
    # print(remove_unecessary(str2))
    # print('-----------------------')
    return remove_unecessary(str1) == remove_unecessary(str2)


def sv_files_equal(fn1, fn2):
    with open(fn1, 'r') as f1:
        with open(fn2, 'r') as f2:
            return equal_on_nonspace(f1.read(), f2.read())


def get_cur_test_name():
    for _, filename, _, function_name, _, _ in inspect.stack():
        if function_name.startswith('test_'):
            return os.path.splitext(filename)[0], function_name
    else:
        raise Exception("Has to be run from within a test function")


def get_result_dir(filename=None, function_name=None):
    if not filename:
        filename, function_name = get_cur_test_name()

    test_dir = os.path.dirname(__file__)

    return os.path.join(
        test_dir, 'result', os.path.relpath(filename, test_dir), function_name)


def prepare_result_dir(filename=None, function_name=None):
    res_dir = get_result_dir(filename, function_name)
    try:
        shutil.rmtree(res_dir)
    except FileNotFoundError:
        pass

    os.makedirs(res_dir, exist_ok=True)

    return res_dir


def get_sv_file_comparison_pair(fn, filename=None, function_name=None):
    if not filename:
        filename, function_name = get_cur_test_name()

    res_dir = get_result_dir(filename, function_name)
    return os.path.join(filename, function_name, fn), os.path.join(res_dir, fn)


def get_test_res_ref_dir_pair(func):
    filename = os.path.splitext(os.path.abspath(inspect.getfile(func)))[0]

    outdir = prepare_result_dir(filename, func.__name__)

    return filename, outdir


def formal_check(disable=None, asserts=None, assumes=None, **kwds):
    def decorator(func):
        return pytest.mark.usefixtures('formal_check_fixt')(
            pytest.mark.parametrize(
                'formal_check_fixt', [[disable, asserts, assumes, kwds]],
                indirect=True)(func))

    return decorator


@pytest.fixture
def formal_check_fixt(tmpdir, request):
    skip_ifndef('FORMAL_TEST')
    yield

    outdir = tmpdir
    disable = request.param[0] if request.param[0] is not None else {}
    asserts = request.param[1] if request.param[1] is not None else {}
    assumes = request.param[2] if request.param[2] is not None else []
    reg['vgen/formal/asserts'] = asserts
    reg['vgen/formal/assumes'] = assumes

    root = find('/')
    module = hdlgen(
        root.child[0], lang='v', outdir=outdir, wrapper=False, **request.param[3])

    yosis_cmds = []
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(searchpath=os.path.dirname(__file__)))
    jinja_context = {'name': module.basename, 'outdir': outdir}

    def find_yosis_cmd(name):
        if name in disable:
            if disable[name] == 'all':
                return
            if disable[name] == 'live':
                jinja_context['live_task'] = False
        script_path = f'{outdir}/top_{name}.sby'
        jinja_context['if_name'] = name.upper()
        env.get_template('formal.j2').stream(jinja_context).dump(script_path)
        yosis_cmds.append(f'sby {script_path}')

    for port in module.in_ports:
        jinja_context['live_task'] = True
        find_yosis_cmd(port.basename)

    for port in module.out_ports:
        jinja_context['live_task'] = False
        find_yosis_cmd(port.basename)

    for cmd in yosis_cmds:
        assert os.system(cmd) == 0, f'Yosis failed. Cmd: {cmd}'


def synth_check(expected, tool='yosys', top=None, **kwds):
    def decorator(func):
        return pytest.mark.usefixtures('synth_check_fixt')(
            pytest.mark.parametrize(
                'synth_check_fixt', [[expected, kwds, tool, top]], indirect=True)(func))

    return decorator


@pytest.fixture
def synth_check_fixt(tmpdir, lang, request):
    skip_ifndef('SYNTH_TEST')

    # lang = 'sv'
    util_ref = request.param[0]
    params = request.param[1]
    tool = request.param[2]
    top = request.param[3]

    if tool == 'vivado':
        skip_ifndef('SYNTH_TEST')

    if tool == 'vivado':
        if not shutil.which('vivado'):
            raise unittest.SkipTest(f"Skipping test, vivado not found")

        params['util'] = True
        params['timing'] = True

    elif tool == 'yosys' and lang == 'v':
        if lang != 'v':
            raise unittest.SkipTest(f"Skipping test, unsupported lang for yosys")

        if not shutil.which('yosys'):
            raise unittest.SkipTest(f"Skipping test, yosys not found")

        params['synthcmd'] = 'synth_xilinx'

        # from pygears.hdl.yosys import synth
        # report = synth(tmpdir, top=top, synth_cmd='synth_xilinx', optimize=True)
    else:
        raise unittest.SkipTest(f"Skipping test, not appropriate tool not found")

    yield

    if top is None:
        top = find('/').child[0]

    util = synth(tool, outdir=tmpdir, top=top, lang=lang, **params)

    if tool == 'vivado':
        util = util['util']

    print(util)

    for param, value in util_ref.items():
        if callable(value):
            assert value(util[param])
        else:
            assert util[param] == value


def hdl_check(expected, **kwds):
    def decorator(func):
        return pytest.mark.usefixtures('hdl_check_fixt')(
            pytest.mark.parametrize('hdl_check_fixt', [[expected, kwds]],
                                    indirect=True)(func))

    return decorator


clear = pytest.fixture(autouse=True)(clear)


@pytest.fixture
def hdl_check_fixt(tmpdir, request):
    reg['gear/infer_signal_names'] = True

    yield

    lang = os.path.splitext(request.param[0][0])[1][1:]
    register_hdl_paths(tmpdir)
    hdlgen(lang=lang, outdir=tmpdir, **request.param[1])

    for fn in request.param[0]:
        res_file = os.path.join(tmpdir, fn)
        ref_file = os.path.join(
            os.path.splitext(request.fspath)[0], request.function.__name__, fn)

        assert sv_files_equal(res_file, ref_file)


def skip_ifndef(*envars):
    import unittest
    import os
    if any(v not in os.environ for v in envars):
        raise unittest.SkipTest(f"Skipping test, {envars} not defined")


def skip_sim_if_no_tools():
    if ('VERILATOR_ROOT' not in os.environ) or ('SYSTEMC_HOME' not in os.environ) or (
            'SCV_HOME' not in os.environ):
        raise unittest.SkipTest("Such-and-such failed. Skipping all tests in foo.py")


@pytest.fixture(
    params=[
        None,
        partial(SimVerilated, lang='v'),
        partial(SimVerilated, lang='sv'),
        SimSocket,
    ])
def sim_cls(request):
    sim_cls = request.param
    if sim_cls is SimVerilated:
        skip_ifndef('VERILATOR_ROOT')
    elif sim_cls is SimSocket:
        skip_ifndef('SIM_SOCKET_TEST')
        sim_cls = partial(SimSocket, run=True, sim='xsim')

    yield sim_cls


@pytest.fixture(
    params=[
        partial(SimVerilated, lang='v'),
        partial(SimVerilated, lang='sv'),
        SimSocket,
    ])
def cosim_cls(request):
    cosim_cls = request.param
    if cosim_cls is SimVerilated:
        skip_ifndef('VERILATOR_ROOT')
    elif cosim_cls is SimSocket:
        skip_ifndef('SIM_SOCKET_TEST')
        cosim_cls = partial(SimSocket, run=True, sim='xsim')

    yield cosim_cls


@pytest.fixture(params=['v', 'sv'])
def lang(request):
    lang = request.param
    # if lang is 'v':
    #     skip_ifndef('VERILOG_TEST')
    yield lang


from pygears import gear
from pygears.lib import decouple


def get_decoupled_dut(delay, f):
    if delay > 0:
        return f

    @gear
    def decoupled(*din):
        return din | f | decouple

    return decoupled



from ..ir_utils import IrExprRewriter, IrRewriter
from pygears.core.datagear import is_datagear


class CallExprRewriter(IrExprRewriter):
    def __init__(self, ctx):
        self.ctx = ctx

    def visit_CallExpr(self, expr):
        if is_datagear(expr.func):
            breakpoint()
            # return call_datagear(expr.func, expr.args, expr.params, self.ctx)
        else:
            return expr


class ExprFinder(IrRewriter):
    def __init__(self, ctx):
        self.ctx = ctx
        super().__init__()

    def Expr(self, expr):
        return CallExprRewriter(self.ctx).visit(expr)


def resolve_gear_calls(modblock, ctx):
    modblock = ExprFinder(ctx).visit(modblock)

    return modblock



import itertools
import os
import socket

from pygears import GearDone, reg
from pygears.conf import Inject, inject
from pygears.sim import clk
from pygears.sim.modules.cosim_base import CosimBase, CosimNoData
from pygears.sim.extens.svsock import register_intf
from pygears.hdl import hdlgen, list_hdl_files
from pygears.hdl.sv.util import svgen_typedef
from pygears.hdl.templenv import TemplateEnv

CMD_SYS_RESET = 0x80000000
CMD_SET_DATA = 0x40000000
CMD_RESET = 0x20000000
CMD_FORWARD = 0x10000000
CMD_CYCLE = 0x08000000
CMD_READ = 0x04000000
CMD_ACK = 0x02000000


class CosimulatorStartError(Exception):
    pass


def format_list(list_, pattern):
    return [pattern.format(s) for s in list_]


async def drive_reset(duration):
    simsoc = reg['sim/config/socket']
    await clk()
    simsoc.send_cmd(duration | CMD_SYS_RESET)
    for i in range(duration):
        await clk()

    # return data


class SimSocketDrv:
    @inject
    def __init__(self, port, index, main=Inject('sim/svsock/server')):
        self.main = main
        self.index = index
        self.port = port

    def reset(self):
        pass


class SimSocketInputDrv(SimSocketDrv):
    def send(self, data):
        # print(
        #     f'{timestep()} [{self.port.name}] Sending {hex(data.code())}, {repr(data)}'
        # )
        self.main.send_cmd(CMD_SET_DATA | self.index)
        self.main.dtype_send(data, self.port.dtype)

    def reset(self):
        # print(f'{timestep()} [{self.port.name}] Reset valid')
        self.main.send_cmd(CMD_RESET | self.index)

    def ready(self):
        self.main.send_cmd(CMD_READ | self.index)

        data = self.main.recv(4)
        res = bool(int.from_bytes(data, byteorder='little'))

        # print(f'{timestep()} [{self.port.name}] Ready={res}')
        return res


class SimSocketOutputDrv(SimSocketDrv):
    def read(self):
        # print(
        #     f'{timestep()} [{self.port.name}] Send read command for {self.index}'
        # )
        self.main.send_cmd(CMD_READ | self.index)
        data = self.main.recv(4)
        # print(
        #     f'{timestep()} [{self.port.name}] Received valid status for {self.index}: {data}'
        # )

        if int.from_bytes(data, byteorder='little'):
            return self.main.dtype_recv(self.port.dtype)
        else:
            raise CosimNoData

    def reset(self):
        self.main.send_cmd(CMD_RESET | self.index)

    def ack(self):
        try:
            self.main.send_cmd(CMD_ACK | self.index)
        except socket.error:
            raise GearDone


class SVServerModule:
    def __init__(self, module, tenv, srcdir, rst):
        self.srcdir = srcdir
        self.module = module
        self.tenv = tenv
        self.rst = rst
        from pygears import reg
        self.svmod = reg['hdlgen/map'][self.module]

    def files(self):
        return list_hdl_files(self.module, self.srcdir)

    def includes(self):
        return reg[f'svgen/include'] + [self.srcdir]

    def declaration(self):
        port_map = {
            port.basename: port.basename
            for port in itertools.chain(self.module.in_ports, self.module.out_ports)
        }
        return self.tenv.snippets.module_inst(
            self.svmod.wrap_module_name, self.svmod.params, "dut", port_map, self.rst)


class SVServerIntf:
    def __init__(self, port, tenv):
        self.port = port
        self.name = port.basename
        self.dtype = port.dtype
        self.tenv = tenv

    def declaration(self):
        return '\n'.join(
            [
                svgen_typedef(self.dtype, f"{self.name}"),
                f"dti_verif_if#({self.name}_t) {self.name}_vif (clk, rst);",
                f"dti#({self.dtype.width}) {self.name} ();",
                f"bit[{self.dtype.width-1}:0] {self.name}_data;",
                self.tenv.snippets.connect_vif(self.name, self.port.direction),
            ])

    def init(self):
        if self.port.direction == "in":
            return '\n'.join(
                [
                    f'{self.name}_vif.name = "{self.name}";',
                    f"{self.name}_vif.valid <= 1'b0;",
                ])
        else:
            return '\n'.join(
                [
                    f'{self.name}_vif.name = "{self.name}";',
                    f"{self.name}_vif.ready <= 1'b0;",
                ])

    def set_data(self):
        if self.port.direction == 'in':
            return self.tenv.snippets.set_data(self.name, self.dtype.width)

    def read(self):
        return self.tenv.snippets.read(self.name, self.port.direction)

    def ack(self):
        if self.port.direction == 'out':
            return self.tenv.snippets.ack(self.name)

    def reset(self):
        return self.tenv.snippets.reset(self.name, self.port.direction)

    def sys_reset(self):
        return self.tenv.snippets.sys_reset(self.name, self.port.direction)


class SimSocket(CosimBase):
    def __init__(self, gear, timeout=100, rebuild=True, run=True, batch=True, rst=True, **kwds):
        super().__init__(gear, timeout)
        self.name = gear.name[1:].replace('/', '_')
        self.outdir = os.path.abspath(os.path.join(reg['results-dir'], self.name))

        self.rebuild = rebuild

        reg['sim/svsock/run'] = run

        if not kwds.get('gui', False):
            kwds['batch'] = batch

        self.rst = rst
        self.kwds = kwds
        self.sock = None
        self.cosim_pid = None

        self.handlers = {}

        reg['sim/config/socket'] = self

        self.srcdir = os.path.join(self.outdir, 'src_gen')
        self.rtl_node = hdlgen(gear, outdir=self.srcdir, lang='sv')
        self.svmod = reg['hdlgen/map'][self.rtl_node]

    def cycle(self):
        self.send_cmd(CMD_CYCLE)

    def forward(self):
        self.send_cmd(CMD_FORWARD)

    back = forward

    def setup(self):
        basedir = os.path.dirname(__file__)
        tenv = TemplateEnv(basedir)
        tenv.snippets = tenv.load(basedir, 'svsock_intf.j2').module

        for cp in self.in_cosim_ports:
            sock_id = register_intf(SVServerIntf(cp.port, tenv))
            self.handlers[cp.port.basename] = SimSocketInputDrv(cp.port, sock_id)

        for p in self.gear.out_ports:
            sock_id = register_intf(SVServerIntf(p, tenv))
            self.handlers[p.basename] = SimSocketOutputDrv(p, sock_id)

        register_intf(SVServerModule(self.rtl_node, tenv, self.srcdir, rst=self.rst))

        self.conn = reg['sim/svsock/server']
        self.send_cmd = self.conn.send_cmd

        super().setup()



"""Represents a generic heterogeneous container type akin to records and structs
in other HDLs.

A concrete :class:`Tuple` type is created by describing the fields it consists
of. There are two ways of creating a concrete :class:`Tuple` type: with or
without specifying the names of the fields. If field names are unimportant, the
type is created by listing the field types in square brackets. For an example,
a 2D Point type, where each coordinate is represented by an 8 bit unsigned
integer (:class:`Uint`) can be specified like this::

    Point = Tuple[Uint[8], Uint[8]]

This is same as if the tuple of types were supplied::

    fields = (Uint[8], Uint[8])
    Point = Tuple[fields]

The ``Point`` type is now represented by a :class:`Tuple` with two fields. If
we want to name the fields, we need to specify the fields in form of the Python
dict, with keys as field names and values as field types::

    Point = Tuple[{
        'x': Uint[8],
        'y': Uint[8]
    }]

If we want to create a type that is generic in terms of the field types, we can
provide a placeholder name instead of the concrete type::

    Point = Tuple[{
        'x': 'coord_t',
        'y': 'coord_t'
    }]

``Point`` is now a generic template type, represented by a :class:`Tuple` with
two fields of the same type, which is not yet specified. The ``Point`` template
has a single parameter ``coord_t``, which needs to be specified to get a
concrete type. So if we want to obtain a ``Point`` with 8 bit coordinates, we
can write::

    PointU8 = Point[Uint[8]]

We can also be explicit which template parameter is assigned a concrete type::

    PointU8 = Point[{'coord_t': Uint[8]}]

Tuples can have any other types for their fields, meaning they can be nested
to create more complex structures::

    Line = Tuple[Point, Point]
    LineU8 = Line[Uint[8]]

>>> LineU8
Tuple[Tuple[{'x': Uint[8], 'y': Uint[8]}], Tuple[{'x': Uint[8], 'y': Uint[8]}]]

The above example shows that the set of template parameters of a :class:`Tuple`
template type is the union of the sets of template parameters of all
:class:`Tuple` fields. Since all ``Line`` template fields have the same
template parameter ``coord_t``, ``Line`` itself has only one template
parameter, namely ``coord_t``:

>>> Line.templates
['coord_t']

If this behaviour is not desired, template arguments can also be renamed by
specifying a string instead of the concrete type for the template parameter::

    Line = Tuple[Point['coord1_t'], Point['coord2_t']]
    LineU8_U16 = Line[Uint[8], Uint[16]]

>>> Line
Tuple[Tuple[{'x': 'coord1_t', 'y': 'coord1_t'}], Tuple[{'x': 'coord2_t', 'y': 'coord2_t'}]]
>>> LineU8_U16
Tuple[Tuple[{'x': Uint[8], 'y': Uint[8]}], Tuple[{'x': Uint[16], 'y': Uint[16]}]]

Once a concrete type has been formed it can be instantiated, which is useful
for the verification. Type instance is obtained by specifying the values for
the :class:`Tuple` fields in parenthesis, grouped in the Python tuple (can be
any iterable really)::

    ab = LineU8_U16(((0, 0), (1, 1)))

>>> ab[0]
(Uint[8](0), Uint[8](0))
>>> ab[0]['x']
Uint[8](0)
"""

from .base import EnumerableGenericMeta, type_str, type_repr, is_type, typeof
from .base import TemplatedTypeUnspecified
from .base import class_and_instance_method


class TupleType(EnumerableGenericMeta):
    """Implements the :class:`Tuple` generic type.

    All operations on the :class:`Tuple` type are implemented here in the
    :class:`TupleType` class. Operations on the :class:`Tuple` type instances
    are defined in the :class:`Tuple` class.
    """
    def __new__(cls, name, bases, namespace, args=None):
        cls = super().__new__(cls, name, bases, namespace, args)

        args = cls.args
        if not args:
            # Generic parameter values have not been supplied
            return cls
        else:
            cls._args = args
            return cls

    def without(self, *index):
        return Tuple[{k: v for k, v in zip(self.fields, self.args) if k not in index}]

    def __add__(self, other):
        """Combines the fields of two :class:`Tuple` types.

        >>> Tuple[{'x': Uint[8], 'y': Uint[8]}] + Tuple[{'x': Uint[8], 'z': Uint[8]}]
        Tuple[{'x': Uint[8], 'y': Uint[8], 'z': Uint[8]}]
        """
        pself = hasattr(self, '__parameters__')
        pother = typeof(other, Tuple) and hasattr(other, '__parameters__')

        if pself and pother:
            return Tuple[{
                **{k: v
                for k, v in zip(self.fields, self.args)},
                **{k: v
                for k, v in zip(other.fields, other.args)}
            }]
        elif not (pself or pother):
            if typeof(other, Tuple):
                return Tuple[self.args + other.args]
            else:
                return Tuple[self.args + tuple(t for t in other)]
        else:
            raise TypeError(f'Currently not supported to combine named and unnamed tuples')

    def __radd__(self, other):
        """Combines the fields of two :class:`Tuple` types.

        >>> Tuple[{'x': Uint[8], 'y': Uint[8]}] + Tuple[{'x': Uint[8], 'z': Uint[8]}]
        Tuple[{'x': Uint[8], 'y': Uint[8], 'z': Uint[8]}]
        """
        pself = hasattr(self, '__parameters__')
        pother = typeof(other, Tuple) and hasattr(other, '__parameters__')

        if pself and pother:
            return Tuple[{
                **{k: v
                for k, v in zip(other.fields, other.args)}
                **{k: v
                for k, v in zip(self.fields, self.args)},
            }]
        elif not (pself or pother):
            if typeof(other, Tuple):
                return Tuple[other.args + self.args]
            else:
                return Tuple[tuple(t for t in other) + self.args]
        else:
            raise TypeError(f'Currently not supported to combine named and unnamed tuples')


    def __mul__(self, other):
        """Doubles the fields of :class:`Tuple` type.

        >>> Tuple[Uint[2], Uint[4]] * 2
        Tuple[Uint[2], Uint[4], Uint[2], Uint[4]]
        """
        return Tuple[self.args * int(other)]

    def __repr__(self):
        if not self.args or not hasattr(self, '__parameters__'):
            return super().__repr__()
        else:
            return 'Tuple[{%s}]' % ', '.join(
                [f'{repr(f)}: {type_repr(a)}' for f, a in zip(self.fields, self.args)])

    def get(self, name, default=None):
        """Calls :py:meth:`__getitem__` and returns the ``default`` value if it
        fails.

        Args:
            default: Type to return if :py:meth:`__getitem__` fails
        """
        try:
            return self[name]
        except KeyError:
            return default

    def __getitem__(self, key):
        """Get the type of the field or fields specified by the ``key``.

        ::

            Point3 = Tuple[{'x': Uint[8], 'y': Uint[8], 'z': Uint[16]}]

        The key can be a name of the field:

        >>> Point3['x']
        Uint[8]

        The key can be a number that represents the index of the field within
        the :class:`Tuple`:

        >>> Point3[2]
        Uint[16]

        Negative keys are accepted to index from the end of the :class:`Tuple`:

        >>> Point3[-1]
        Uint[16]

        Slices are accepted to return a new :class:`Tuple` with a subset of
        fields:

        >>> Point3[:2]
        Tuple[Uint[8], Uint[8]]

        The key can be a sequence of the names, number indexes or slices, where
        a new :class:`Tuple` is return with a subset of fields given by the
        keys in the sequence:

        >>> Point3['x', -1]
        Tuple[Uint[8], Uint[16]]

        """
        if not self.specified:
            return super().__getitem__(key)

        key_norm = self.index_norm(key)

        if (len(key_norm) == 1) and (not isinstance(key_norm[0], slice)):
            return self.__args__[key_norm[0]]
        else:
            subtypes = []
            for i in key_norm:
                subt = self.__args__[i]
                subtypes.extend(subt if isinstance(i, slice) else [subt])

            return Tuple[tuple(subtypes)]

    def __str__(self):
        return '(%s)' % ', '.join([type_str(a) for a in self.args])

    @property
    def width(self):
        if not self.specified:
            raise TemplatedTypeUnspecified(
                f'Cannot callculate width of the unspecified type {repr(self)}')
        w = 0
        for f in self.__args__:
            if not is_type(f):
                raise TypeError(f'Argument "{repr(f)}" of type "{repr(self)}" is not PyGears type')

            w += f.width

        return w


class Tuple(tuple, metaclass=TupleType):
    """Implements the :class:`Tuple` type instance.

    The :class:`Tuple` type instance can be initialized either with a dict that
    maps the field names to the desired values::

        Point = Tuple[{'x': Uint[8], 'y': Uint[8]}]

    >>> Point({'y': 0, 'x': 1})
    (Uint[8](1), Uint[8](0))

    or with an iterable which enumerates values for all the :class:`Tuple`
    fields in order:

    >>> Point((1, 0))
    (Uint[8](1), Uint[8](0))

    """
    def __new__(cls, val=None):
        if isinstance(val, cls):
            return val

        if not cls.specified:
            try:
                if isinstance(val, dict):
                    cls = Tuple[{n: type(v) for n, v in val.items()}]
                else:
                    cls = Tuple[tuple([type(v) for v in val])]
            except TypeError as e:
                raise TypeError(f"{str(e)}\n - when creating Tuple from '{val}'")

        if val is None:
            tpl_val = tuple(t() for t in cls.__args__)
        elif isinstance(val, dict):
            tpl_val = []
            for t, f in zip(cls.__args__, cls.fields):
                try:
                    tpl_val.append(t(val[f]))
                except (TypeError, ValueError) as e:
                    msg = (f'{str(e)}\n - when instantiating field "{f}" of'
                           f' type "{repr(t)}" with "{repr(val[f])}"')

                    if is_type(val[f]):
                        msg += (f'\n FIX: Did you mean to define a Tuple type?'
                                f' Use square brackets [] instead of ()')

                    raise TypeError(msg)

        else:
            tpl_val = []
            for i, (t, v) in enumerate(zip(cls.__args__, val)):
                try:
                    tpl_val.append(t(v))
                except TypeError as e:
                    msg = (f'{str(e)}\n - when instantiating field {i} of'
                           f' type "{repr(t)}" with "{repr(v)}"')

                    if is_type(v):
                        msg += (f'\n FIX: Did you mean to define a Tuple type?'
                                f' Use square brackets [] instead of ()')

                    raise TypeError(msg)

        if len(tpl_val) != len(cls.__args__):
            raise TypeError(f'{repr(cls)}() takes {len(cls)} arguments' f' ({len(tpl_val)} given)')

        return super(Tuple, cls).__new__(cls, tpl_val)

    @class_and_instance_method
    def __getitem__(self, key):
        """Returns the value of the field or fields specified by the ``key``.

        ::

            Point3 = Tuple[{'x': Uint[8], 'y': Uint[8], 'z': Uint[16]}]
            point_a = Point3((1, 2, 3))

        The key can be a name of the field:

        >>> point_a['x']
        Uint[8](1)

        The key can be a number that represents the index of the field within
        the :class:`Tuple`:

        >>> point_a[2]
        Uint[16](3)

        Negative keys are accepted to index from the end of the :class:`Tuple`:

        >>> point_a[-1]
        Uint[16](3)

        Slices are accepted to return a new :class:`Tuple` with a subset of
        fields:

        >>> point_a[:2]
        (Uint[8](1), Uint[8](2))

        The key can be a sequence of the names, number indexes or slices, where
        a new :class:`Tuple` is return with a subset of fields given by the
        keys in the sequence:

        >>> point_a['x', -1]
        (Uint[8](1), Uint[16](3))
        """

        if isinstance(key, int):
            return super().__getitem__(key)
        elif isinstance(key, str):
            try:
                return super().__getitem__(type(self).fields.index(key))
            except ValueError:
                raise TypeError(f'Tuple "{repr(self)}" has no field "{key}"')

        key_norm = type(self).index_norm(key)

        if (len(key_norm) == 1) and (not isinstance(key_norm[0], slice)):
            return super(Tuple, self).__getitem__(key_norm[0])
        else:
            tout = type(self)[key_norm]
            subtypes = []
            for i in key_norm:
                subt = super(Tuple, self).__getitem__(i)
                subtypes.extend(subt if isinstance(i, slice) else [subt])

            return tout(tuple(subtypes))

    def __add__(self, other):
        return (type(self) + type(other))(super().__add__(tuple(other)))

    __radd__ = __add__

    def __mul__(self, other):
        return (type(self) * int(other))(super().__mul__(int(other)))

    @class_and_instance_method
    def __str__(self):
        return '(%s)' % ', '.join([type_str(a) for a in self])

    def __eq__(self, other):
        if not is_type(type(other)):
            return super().__eq__(other)

        return type(self) == type(other) and super().__eq__(other)

    def __hash__(self):
        return super().__hash__()

    def __ne__(self, other):
        if not is_type(type(other)):
            return super().__ne__(other)

        return not self.__eq__(other)

    @class_and_instance_method
    def subs(self, *args, **kwds):
        if args:
            path, val = args
            if isinstance(path, tuple):
                if len(path) > 1:
                    val = self[path[0]].subs(path[1:], val)

                path = path[0]

            if isinstance(path, int):
                path = type(self).fields[path]

            kwds = {path: val}

        map_dict = {f: kwds.get(f, self[f]) for f in type(self).fields}
        return type(self)(map_dict)

    @class_and_instance_method
    def get(self, name, default=None):
        try:
            return self[name]
        except KeyError:
            return default

    def __int__(self):
        """Returns a packed integer representation of the :class:`Tuple` instance.

        ::

            Point = Tuple[{'x': Uint[8], 'y': Uint[8]}]

        >>> int(Point((0xaa, 0xbb)))
        48042
        >>> hex(48042)
        '0xbbaa'
        """
        ret = 0

        t = type(self).__args__
        for i in range(len(t) - 1, -1, -1):
            ret <<= t[i].width
            ret |= super().__getitem__(i).code() & ((1 << t[i].width) - 1)

        return ret

    code = __int__

    @classmethod
    def decode(cls, val):
        """Returns a :class:`Tuple` instance from its packed integer representation.

        ::

            Point = Tuple[{'x': Uint[8], 'y': Uint[8]}]

        >>> Point.decode(0xbbaa)
        (Uint[8](170), Uint[8](187))
        """
        ret = []
        val = int(val)
        for t in cls:
            t_width = t.width
            t_mask = (1 << t_width) - 1
            ret.append(t.decode(val & t_mask))
            val >>= t_width

        return cls(tuple(ret))



import inspect
import weakref

from pygears import reg, Intf, module, find
from pygears.sim.sim_gear import SimGear, is_simgear_func
from pygears.sim.sim import SimPlugin, cosim
from pygears.core.gear import GearPlugin
from pygears.core.gear_inst import gear_base_resolver
from pygears.core.hier_node import HierVisitorBase
from pygears.core.port import HDLConsumer, HDLProducer

intfs = []


def sim_compile_resolver(func, *args, **kwds):
    ctx = reg['gear/exec_context']
    if ctx == 'sim':
        reg['gear/exec_context'] = 'compile'

        local_in = []
        for a in args:
            if isinstance(a, Intf):
                # a.consumers.clear()
                local_in.append(a)
            else:
                from pygears.lib.const import get_literal_type
                local_in.append(Intf(get_literal_type(a)))
                local_in[-1].producer = HDLProducer()

        reg['gear/exec_context'] = 'sim'

        outputs = gear_base_resolver(func, *local_in, **kwds)

        # TODO: Support multiple outputs

        if isinstance(outputs, tuple):
            raise Exception("Not yet supported")

        # outputs.connect(HDLConsumer())
        outputs.consumers.append(HDLConsumer())

        gear_inst = outputs.producer.gear
        gear_inst.trace = None

        def is_async_gen(func):
            return bool(func.__code__.co_flags & inspect.CO_ASYNC_GENERATOR)

        if not is_async_gen(gear_inst.func):
            raise Exception("Not yet supported")

        simulator = reg['sim/simulator']
        cur_sim = reg['gear/current_sim']
        sim_map = reg['sim/map']

        sim_gear = SimGear(gear_inst)

        sim_map[gear_inst] = sim_gear
        cur_sim.child.append(sim_gear)
        sim_gear.parent = cur_sim
        simulator.forward_ready.add(sim_gear)
        sim_gear.setup()
        simulator.tasks[sim_gear] = sim_gear.run()
        simulator.task_data[sim_gear] = None

        import asyncio
        for intf, a in zip(local_in, args):
            if isinstance(a, Intf):
                continue

            intf._in_queue = asyncio.Queue(maxsize=1, loop=reg['sim/simulator'])
            intf.put_nb(a)

        def callback(p):
            # print(f'Out of scope: {p.gear.name}')
            g = p.gear

            # TODO: in_ports can be missing if it was deleted in order to help GC
            # collect the graph
            if hasattr(g, 'in_ports'):
                for i in g.in_ports:
                    if isinstance(i.producer.producer, HDLProducer):
                        i.producer.finish()
                    else:
                        i.producer.disconnect(i)

        weakref.finalize(outputs, callback, outputs.producer)

        # if not intfs:
        #     intfs.append(outputs)

        # print(f'Referrers:')
        # import gc
        # bla = gc.get_referrers(intfs[0])
        # for b in bla:
        #     print(b)
        # print(f'-----------------------------------')

        return outputs
    else:
        return gear_base_resolver(func, *args, **kwds)


class SimInstVisitor(HierVisitorBase):
    def __init__(self):
        self.namespace = reg['sim/module_namespace']
        self.sim_map = reg['sim/map']

    def Gear(self, module):
        if module.parent is None:
            return False

        if module.params.get('__sim__', None) is not None:
            cfg = module.params['__sim__']
            if isinstance(cfg, dict):
                cfg = cfg.copy()
                sim = cfg.pop('sim')
                cosim(module, sim, **cfg)
            else:
                cosim(module, cfg)

        sim_cls = module.params.get('sim_cls', None)
        sim_inst = None

        if sim_cls is None:
            sim_cls = self.namespace.get(module.definition, None)

        if sim_cls:
            sim_inst = sim_cls(module)
        elif is_simgear_func(module.func):
            sim_inst = SimGear(module)

        if sim_inst:
            self.sim_map[module] = sim_inst
            return True


def sim_inst(top=None):
    if top is None:
        top = find('/')

    v = SimInstVisitor()
    v.visit(top)

    return top


class SimInstPlugin(SimPlugin, GearPlugin):
    @classmethod
    def bind(cls):
        reg['sim/flow'].append(sim_inst)
        reg['sim/module_namespace'] = {}
        reg['sim/map'] = {}
        reg['gear/params/extra/sim_cls'] = None
        reg['gear/params/extra/__sim__'] = None
        reg['gear/gear_dflt_resolver'] = sim_compile_resolver

    @classmethod
    def reset(cls):
        reg['sim/map'] = {}



from pygears import gear, datagear
from pygears.lib import ccat

# @gear
# def repack(cfg, *, t, name_map={}, val_map={}):

#     mapped_names = [name_map.get(f, f) for f in dtype.fields]
#     map_dict = {n: val.get(n, val_map.get(n, None)) for n in mapped_names}

#     return dtype(map_dict)

#     fields = (cfg[name_map.get(f, f)] for f in t.fields)
#     return ccat(*fields)


# @gear
# def repack(cfg, *, t, name_map={}, val_map={}):
#     fields = []
#     for f_name in t.fields:
#         if f_name in val_map:
#             f = val_map[f_name]
#         else:
#             f_name = name_map.get(f_name, f_name)
#             f = cfg[f_name]

#         fields.append(f)

#     return ccat(*fields) | t

# def repack_func(t, name_map, val_map):
#     fields = []
#     for f_name in t.fields:
#         if f_name in val_map:
#             f = val_map[f_name]
#         else:
#             f_name = name_map.get(f_name, f_name)
#             f = cfg[f_name]

#         fields.append(f)

#     return tuple(fields)

@datagear
def repack(cfg, *, t, name_map={}, val_map={}) -> b't':
    return t([val_map[n] if n in val_map else cfg[name_map.get(n, n)] for n in t.fields])



try:
    import matplotlib
    from matplotlib.ticker import FuncFormatter
    import matplotlib.pyplot as plt
    import matplotlib.animation as animation
except ImportError:
    pass

import os
import queue
import pickle
from functools import partial
from pygears.typing import Array, typeof, Queue
from pygears import gear, module, GearDone, reg
from pygears.sim import timestep, clk

import multiprocessing

suffix = ["G", "M", "k", "", "m", "u", "n", "p"]
decades = [1e9, 1e6, 1e3, 1e0, 1e-3, 1e-6, 1e-9, 1e-12]

colors = {
    'line': "#355c7d",
    'handshake': "#c06c84",
    'eot': "#f67280",
}


def t_fmt(t, posp, positions):
    if t == 0:
        return str(0)

    for i, d in enumerate(decades):
        if abs(t) >= d:
            val = int(round((t / float(d)) * (10**positions)))
            str_val = str(val)
            str_val = str_val[:-positions] + '.' + str_val[-positions:]

            for cutoff in range(0, positions + 2):
                if str_val[-(cutoff + 1)] in ['0', '.']:
                    continue

                break

            if cutoff:
                str_val = str_val[:-cutoff]

            return f'{str_val} {suffix[i]}'

    return t


def create_plot(title, method, clk_freq):
    fig, ax = plt.subplots()

    if title:
        plt.title(title)

    line = []
    for m in method:
        if isinstance(m, str):
            kwds = {'linewidth': 2, 'color': colors['line']}
            if m == 'step':
                kwds['where'] = 'post'

            line.append(getattr(ax, m)([], [], **kwds)[0])
        else:
            line.append(m(ax, x=[], y=[]))

    ax.grid()
    ax.set_xlim(0, 1 / clk_freq)

    tmp = clk_freq
    positions = 0
    while tmp > 1:
        positions += 1
        tmp /= 10

    ffmt = FuncFormatter(partial(t_fmt, positions=positions))
    fig.ffmt = ffmt
    ax.xaxis.set_major_formatter(ffmt)

    return fig, ax, line


def plot_process(method, qin, clk_freq, title=None, scale=None):

    fig, ax, line = create_plot(title, method, clk_freq)

    chnum = len(method)

    xdata = [[] for _ in range(chnum)]
    ydata = [[] for _ in range(chnum)]

    yrng = [0, 0]
    done = []

    class CustFuncAnimation(animation.FuncAnimation):
        def _init_draw(self):
            try:
                super()._init_draw()
            except StopIteration:
                super()._draw_frame(tuple(yrng))

    def data_gen():
        while True:
            while True:
                if done:
                    return

                try:
                    res = qin.get(block=True, timeout=1.0)
                except queue.Empty:
                    yield tuple(yrng)
                    continue

                if res is None:
                    done.append(True)
                    break

                ch, t, y, _ = res

                if scale:
                    y *= scale[ch]

                xdata[ch].append(t)
                ydata[ch].append(y)

                if y > yrng[1]:
                    yrng[1] = y

                if y < yrng[0]:
                    yrng[0] = y

                if qin.empty():
                    break

            yield tuple(yrng)

            if res is None:
                break

    def run(yrng):
        t = max(xdata[i][-1] if xdata[i] else 0 for i in range(chnum))
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()

        redraw = bool(done)
        if t >= xmax:
            ax.set_xlim(xmin, 1.5 * t)
            redraw = True

        if (ymax < yrng[1] * 1.1) or (ymin > yrng[0]):
            ax.set_ylim(yrng[0], yrng[1] * 1.1)
            redraw = True

        for i in range(chnum):
            line[i].set_data(xdata[i], ydata[i])

        if redraw:
            ax.figure.canvas.draw()

        return

    _ = CustFuncAnimation(fig, run, data_gen, blit=False, interval=100, repeat=False)

    plt.show()


def plot_live(method, clk_freq, title, scale, transaction):
    qin = multiprocessing.Queue(maxsize=10000)

    _proc = multiprocessing.context.Process(
        target=plot_process, args=(method, qin, clk_freq, title, scale))
    _proc.start()

    def cleanup(sim):
        if not reg['sim/exception']:
            _proc.join()
        else:
            _proc.terminate()

    reg['sim/simulator'].events['after_cleanup'].append(cleanup)

    try:
        while True:
            qin.put((yield None))
    except GeneratorExit:
        qin.put(None)


def plot_dump(files, method, clk_freq, title, scale, transaction):
    fig, ax, line = create_plot(title, method, clk_freq)

    chnum = len(method)

    xdata = [[] for _ in range(chnum)]
    ydata = [[] for _ in range(chnum)]
    eots = [[] for _ in range(chnum)]

    from matplotlib import patches
    try:
        while True:
            ch, t, val, eot = yield None
            if eot:
                eots[ch].append(t)

            xdata[ch].append(t)
            ydata[ch].append(val)
    except GeneratorExit:
        pass

    ax.set_autoscale_on(True)
    for i in range(chnum):
        ymax = max(ydata[i])
        ymin = min(ydata[i])

        def handshake(x1, x2, color, alpha):
            rect = patches.Rectangle(
                (x1, ymin), x2 - x1, ymax - ymin, color=color, fill=True, alpha=alpha)
            ax.add_patch(rect)

        if transaction:
            next_t = None
            rect_start_t = None

            for t in xdata[i]:
                if rect_start_t is None:
                    rect_start_t = t

                if (next_t is not None and (t - next_t) > 1 / (2 * clk_freq)):
                    handshake(rect_start_t, next_t, colors['handshake'], 0.13)
                    rect_start_t = t

                next_t = t + 1 / clk_freq

            if (next_t is not None and (next_t - rect_start_t) > 1 / (2 * clk_freq)):
                handshake(rect_start_t, next_t, colors['handshake'], 0.13)

            for t in eots[i]:
                handshake(t, t + 1 / clk_freq, colors['eot'], 0.5)

        ydata[i].append(ydata[i][-1])
        xdata[i].append(xdata[i][-1] + 1 / clk_freq)
        line[i].set_data(xdata[i], ydata[i])

    ax.relim()
    ax.autoscale_view(True, True, True)
    # plt.axhline(xmin=0.1, xmax=0.2, linewidth=8, color='#d62728')
    # plt.hlines(y=0.5, xmin=0.0, xmax=0.02, linewidth=8, color='#d62728', solid_capstyle='round')
    # plt.fill_between([0, 0.02], [10, 10])

    # Add the patch to the Axes
    # plt.bar([0, 0.02], [0, 0], linewidth=8, color='#d62728', solid_capstyle='round', alpha=0.5)
    # plt.axvspan(0, 0.02, facecolor='#3cb03c', alpha=0.3)
    # plt.axvspan(0.02, 0.04, facecolor='#0c600c', alpha=0.35)
    # ax.axhline(linewidth=8, color='#d62728')
    fig.canvas.draw()

    for f in files.split(','):
        if f == 'show':
            continue

        os.makedirs(os.path.dirname(f), exist_ok=True)
        if os.path.splitext(f)[1] == '.pkl':
            with open(f, 'wb') as fid:
                pickle.dump(fig, fid)
        else:
            fig.savefig(f)

    if 'show' in files.split(','):
        plt.show()

    plt.close(fig)


@gear(enablement=b'"matplotlib" in globals()')
async def scope(
    *xs,
    clk_freq=None,
    title=None,
    scale=None,
    method=None,
    live=None,
    dump=None,
    transaction=False):

    if clk_freq is None:
        clk_freq = reg['sim/clk_freq']

    if method is None:
        method = ['plot'] * len(xs)

    if len(method) != len(xs):
        raise Exception(
            f'Number of plotting methods ({method}) needs to match the number of inputs ({len(xs)})')

    if title is None:
        title = module().name

    parallel_steps = max(len(x.dtype) if typeof(x.dtype, Array) else 1 for x in xs)

    backends = []

    kwds = {
        'method': method,
        'clk_freq': clk_freq * parallel_steps,
        'title': title,
        'scale': scale,
        'transaction': transaction
    }

    if live or (live is None and dump is None):
        backends.append(plot_live(**kwds))

    if dump:
        backends.append(plot_dump(dump, **kwds))

    for b in backends:
        b.send(None)

    try:
        while True:
            for ch, x in enumerate(xs):
                if x.done:
                    raise GearDone

                if x.empty():
                    continue

                async with x as x_data:
                    if isinstance(x_data, Queue):
                        x_data, eot = x_data
                    else:
                        eot = 0

                    if not isinstance(x_data, Array):
                        x_data = [x_data]

                    for i, v in enumerate(x_data):
                        point = (
                            ch, (timestep() + i / len(x_data)) / clk_freq, float(v), int(eot))

                        for b in backends:
                            b.send(point)

            await clk()

    except GearDone:
        for b in backends:
            b.close()
        raise GearDone



from pygears import gear
from pygears.typing import Array, Tuple, Uint
from pygears.sim import clk


@gear
async def strb_combiner(din: Tuple[{'data': Array['data_t', 'num'], 'strb': Uint}]) -> b'din["data"]':
    data_t = din.dtype['data'].data
    num = len(din.dtype['data'])

    data = Array[data_t, num]()
    last = False

    while not last:
        async with din as (d, strb):
            for i in range(num):
                if strb[i]:
                    data[i] = d[i]

            last = strb[-1]

    await clk()
    yield data


# import sys
# sys.setrecursionlimit(1500)
# # array_t = Array[Uint[8], 4]
# # strb_t = Uint[4]
# # seq = [((1, 2, 3, 4), 0x3), ((5, 6, 7, 8), 0xc)]

# array_t = Array[Uint[8], 64]
# strb_t = Uint[64]
# seq = [(list(range(64)), (1<<32) - 1), (list(range(64, 128)), ((1<<32) - 1) << 32)]


# res = []
# drv(t=Tuple[array_t, strb_t], seq=seq) \
#     | strb_combiner \
#     | collect(result=res)

# cosim('/strb_combiner', 'verilator')
# sim()

# breakpoint()



from pygears.sim import clk, delta, log
from pygears import GearDone


class CosimNoData(Exception):
    pass


class InCosimPort:
    def __init__(self, main, port, name=None):
        self.port = port
        self.name = name if name else port.basename
        self.done = False
        self.parent = None
        self.child = None
        self.main = main

    @property
    def gear(self):
        return self.port.gear

    @property
    def in_ports(self):
        return [self.port]

    @property
    def out_ports(self):
        return []

    def setup(self):
        pass

    async def run(self):
        intf = self.port.consumer

        self.active = False

        while True:
            if self.main.done:
                intf.finish()
                raise GearDone

            try:
                if self.phase == 'forward':
                    if not self.active:
                        self.main.reset_in(self.name)
                        # log.info(f'Wait for intf -> {self.name}')
                        data = await intf.pull()
                        # log.info(f'Set {data} -> {self.name}')
                        self.active = True
                        self.main.write_in(self.name, data)

                    await delta()
                elif self.phase == 'back':
                    if self.active:
                        if self.main.ready_in(self.name):
                            # log.info(f'Ack {self.name}')
                            self.active = False
                            intf.ack()
                        else:
                            pass
                            # log.info(f'NAck {self.name}')
                    await clk()

            except (BrokenPipeError, ConnectionResetError):
                intf.finish()
                raise GearDone

            # phase = await delta()


class OutCosimPort:
    def __init__(self, main, port, name=None):
        self.port = port
        self.name = name if name else port.basename
        self.done = False
        self.parent = None
        self.child = None
        self.main = main

    @property
    def gear(self):
        return self.port.gear

    @property
    def in_ports(self):
        return []

    @property
    def out_ports(self):
        return [self.port]

    def setup(self):
        pass

    async def run(self):
        intf = self.port.producer

        while True:
            if self.main.done:
                # log.info(f'CosimPort {self.name} finished')
                intf.finish()
                raise GearDone

            try:
                # log.info(f'{self.name} read_out')
                data = self.main.read_out(self.name)
                # log.info(f'Put {data} -> {self.name}')
                await intf.put(data)
                # log.info(f'Ack {data} -> {self.name}')
                self.main.ack_out(self.name)
            except (BrokenPipeError, ConnectionResetError):
                intf.finish()
                raise GearDone
            except CosimNoData:
                pass

            await clk()



import inspect

from .base import EnumerableGenericMeta, type_str, typeof
from .base import TemplatedTypeUnspecified, class_and_instance_method, is_type
from .uint import Uint
from .unit import Unit


class QueueMeta(EnumerableGenericMeta):
    _eot = None

    def keys(self):
        return (0, 1)

    @property
    def width(self):
        return sum(f.width for f in self)
        # return self.__args__[0].width + self.__args__[1]

    def __new__(cls, name, bases, namespace, args=None):
        if args:
            if isinstance(args, dict) and (args['eot'] == 0):
                return args['data']

            if isinstance(args, list) and (typeof(args[0], Queue)):
                args = (args[0].data, args[0].lvl + args[1])
            elif isinstance(args, dict) and (typeof(args['data'], Queue)):
                args = (args['data'].data, args['data'].lvl + args['eot'])

        return super().__new__(cls, name, bases, namespace, args)

    def __getitem__(self, index):
        if not self.specified:
            if inspect.isclass(index) and issubclass(index, Queue) and not self.__args__:
                return Queue[index.args[0], index.lvl + 1]
            elif isinstance(index, tuple) and inspect.isclass(index[0]) and issubclass(
                    index[0], Queue) and not self.__args__:
                return Queue[index[0].args[0], index[0].lvl + index[1]]
            elif isinstance(index, tuple) and (index[1] == 0):
                return index[0]
            else:
                return super().__getitem__(index)

        if index == 0:
            return self.args[0]
        elif index == 1:
            return self.eot

        key_norm = self.index_norm(index)
        if (len(key_norm) > 1):
            raise IndexError

        key_norm = key_norm[0]

        if not isinstance(key_norm, slice):
            key_norm = slice(key_norm, key_norm + 1)

        if key_norm.start == key_norm.stop:
            return Unit
        elif key_norm.start == 0:
            if key_norm.stop == 1:
                return self.data
            else:
                return self
        elif key_norm.start == 1:
            return self.eot
        else:
            raise IndexError

    def sub(self, lvl=None):
        if lvl is None:
            lvl = self.lvl - 1

        return Queue[self.data, lvl]

    def wrap(self, lvl=1):
        return Queue[self.data, self.lvl + lvl]

    @property
    def lvl(self):
        return self.args[1]

    @property
    def data(self):
        return self.__args__[0]

    @property
    def eot(self):
        if self._eot is None:
            self._eot = Uint[self.lvl]

        return self._eot

    def __str__(self):
        if self.args:
            if self.lvl == 1:
                return '[%s]' % type_str(self.args[0])
            else:
                return '[{}]^{}'.format(type_str(self.args[0]), self.lvl)
        else:
            return super().__str__()


# TODO: If queue is parameterized with two types, the other one will substitute eot field
class Queue(tuple, metaclass=QueueMeta):
    __default__ = (1, )
    __parameters__ = ['data', 'eot']

    def __new__(cls, val=None, eot=None):
        if type(val) == cls:
            return val

        if not cls.specified:
            raise TemplatedTypeUnspecified

        if val is None and eot is None:
            return super(Queue, cls).__new__(cls, (cls.args[0](), cls.eot()))

        if eot is None:
            val, eot = val

        queue_tpl = (cls.args[0](val), cls.eot(eot))
        return super(Queue, cls).__new__(cls, queue_tpl)

    def __eq__(self, other):
        if not is_type(type(other)):
            return super().__eq__(other)

        return type(self) == type(other) and super().__eq__(other)

    def __hash__(self):
        return super().__hash__()

    def __ne__(self, other):
        if not is_type(type(other)):
            return super().__ne__(other)

        return not self.__eq__(other)

    def __int__(self):
        """Returns a packed integer representation of the :class:`Queue` instance.
        """
        ret = 0

        for d, t in zip(reversed(self), reversed(type(self))):
            ret <<= t.width
            ret |= int(d)

        return int(self.data) | (int(self.eot) << type(self).data.width)

    def code(self):
        """Returns a packed integer representation of the :class:`Queue` instance.
        """
        data = super().__getitem__(0)
        eot = super().__getitem__(1)
        return data.code() | (int(eot) << type(data).width)

    @class_and_instance_method
    def sub(self, lvl=None):
        cls_ret = type(self).sub(lvl)
        if typeof(cls_ret, Queue):
            return type(self).sub(lvl)(self.data, self.eot[:cls_ret.lvl])
        else:
            return self.data

    @class_and_instance_method
    def wrap(self, eot):
        if not isinstance(eot, Uint):
            eot = Uint(eot)

        wrap_cls = type(self).wrap(eot.width)
        wrap_eot = wrap_cls.eot(self.eot) | (eot << self.lvl)

        return wrap_cls(self.data, wrap_eot)

    @property
    def last(self):
        return all(self.eot)

    @class_and_instance_method
    @property
    def eot(self):
        return self[1]

    @class_and_instance_method
    @property
    def data(self):
        return self[0]

    @class_and_instance_method
    @property
    def lvl(self):
        return type(self).lvl

    @classmethod
    def decode(cls, val):
        data_t = cls.__args__[0]
        data_t_width = data_t.width
        data_t_mask = (1 << data_t_width) - 1
        val = int(val)

        return cls((data_t.decode(val & data_t_mask), val >> data_t_width))




from ..cfg import CfgDfs, Node
from ..ir_utils import ir
from ..cfg_util import remove_node


def cond_wrap(first, last, test):
    source = first.prev
    sink = last.next

    if source[0] is last:
        return None

    cond_blk = Node(ir.HDLBlock(), prev=source)
    Node(ir.HDLBlockSink(), source=cond_blk, next_=sink)

    b_if = Node(ir.Branch(test=test), prev=[cond_blk], next_=[first])
    b_else = Node(ir.Branch(), prev=[cond_blk])
    cond_blk.next = [b_if, b_else]

    Node(ir.BranchSink(), source=b_if, prev=[last], next_=[cond_blk.sink])
    Node(ir.BranchSink(), source=b_else, prev=[b_else], next_=[cond_blk.sink])
    b_else.next = [b_else.sink]
    cond_blk.sink.prev = [b_else.sink, b_if.sink]

    first.prev = [b_if]
    last.next = [b_if.sink]

    i = source[0].next.index(first)
    source[0].next[i] = cond_blk
    sink[0].prev = [cond_blk.sink]

    return cond_blk


# TODO: Remove scoping, it is implemented in CfgDfs
class ResolveBlocking(CfgDfs):
    def __init__(self, ctx):
        self.scopes = []
        self.ctx = ctx

    @property
    def scope(self):
        return self.scopes[-1]

    def apply_await(self, node, blocking):
        if blocking == ir.res_true or not node.next:
            return

        cond_wrap(node.next[0], self.scope.sink.prev[0], blocking)
        self.scope.blocking = ir.BinOpExpr((self.scope.blocking, blocking), ir.opc.And)

    def enter_BaseBlock(self, node):
        self.scopes.append(node)
        node.blocking = ir.res_true
        node.break_ = ir.res_true

    def exit_BaseBlock(self, node):
        self.scopes.pop()
        self.apply_await(node.sink, node.blocking)

    def exit_LoopBody(self, node):
        self.scopes.pop()
        if node.break_ != ir.res_true:
            node.blocking = ir.BinOpExpr([ir.UnaryOpExpr(node.break_, ir.opc.Not), node.blocking],
                                         ir.opc.Or)

        self.apply_await(node.sink, node.blocking)

    def exit_Branch(self, node):
        self.scopes.pop()

    def exit_Await(self, node):
        if node.value.expr not in ['forward', 'back', 'break']:
            self.apply_await(node, node.value.expr)

        remove_node(node)

    def enter_Jump(self, stmt):
        irnode: ir.Jump = stmt.value
        if irnode.label == 'state':
            stmt.value = ir.AssignValue(self.ctx.ref('_state'), ir.ResExpr(irnode.where))
            self.apply_await(stmt, ir.res_false)
        elif irnode.label == 'break':
            self.apply_await(stmt, ir.res_false)
            self.scope.break_ = ir.res_false
            # self.scope.break_ = ir.BinOpExpr((self.scope.break_, self.scope.blocking), ir.opc.And)
            remove_node(stmt)
        else:
            remove_node(stmt)

    def exit_HDLBlock(self, block):
        blocking = ir.res_true
        for b in reversed(block.next):
            blocking = ir.ConditionalExpr((b.blocking, blocking), b.value.test)

        break_ = ir.res_true
        for b in reversed(block.next):
            break_ = ir.ConditionalExpr((b.break_, break_), b.value.test)

        self.scope.break_ = ir.BinOpExpr((self.scope.break_, break_), ir.opc.And)

        # TODO: Rest can be put in else statement if else has no blocking
        # statements. This might result in simpler code
        self.apply_await(block.sink, blocking)



from queue import Queue
import attr
import inspect
from .. import cfg as cfgutil
from contextlib import contextmanager
from copy import deepcopy, copy
from ..cfg import Node, draw_cfg, CfgDfs
from ..cfg_util import insert_node_before, insert_node_after
from ..ir_utils import res_true, HDLVisitor, ir, add_to_list, res_false, is_intf_id, IrRewriter, IrExprVisitor
from pygears.typing import bitw, Uint, Bool
from .inline_cfg import VarScope, ExploreState
from .exit_cond_cfg import ResolveBlocking, cond_wrap
from ..cfg_util import insert_node_before


class VariableFinder(IrExprVisitor):
    def __init__(self):
        self.variables = set()

    def visit_AssignValue(self, node: ir.AssignValue):
        self.visit(node.target)
        self.visit(node.val)

    def visit_Branch(self, node: ir.Branch):
        self.visit(node.test)

    def visit_Name(self, node):
        if node.ctx == 'load':
            self.variables.add(node.name)


def create_scheduled_cfg(node, G, visited, labels, reaching=None, simple=True):
    if node in visited:
        return

    visited.add(node)
    if node.value is None:
        name = "None"
    elif isinstance(node.value, ir.Branch):
        if node.value.test == ir.res_true:
            name = 'else'
        else:
            name = f'if {node.value.test}'
    elif isinstance(node.value, (ir.HDLBlock, ir.BaseBlock)):
        name = f'{type(node.value)}'
    else:
        name = str(node.value)

    if reaching and node in reaching:
        name += ' <- ' + ','.join([str(rin[1]) for rin in reaching[node].get('in', {})])

    labels[id(node)] = name

    for n in node.next:
        if simple:
            if isinstance(n.value, ir.Branch) and n.next:
                n = n.next[0]

            while (n.next and
                   (isinstance(n.value, ir.BaseBlockSink) or isinstance(n.value, ir.HDLBlockSink))):
                n = n.next[0]

        G.add_edge(id(node), id(n))

        create_scheduled_cfg(n, G, visited, labels, reaching=reaching, simple=simple)


def draw_scheduled_cfg(cfg, reaching=None, simple=True):
    import networkx as nx
    import matplotlib.pyplot as plt

    G = nx.DiGraph()
    visited = set()
    labels = {}

    create_scheduled_cfg(cfg, G, visited, labels, reaching=reaching, simple=simple)

    pos = nx.planar_layout(G)
    nx.draw(G, pos, font_size=16, with_labels=False)

    # for p in pos:  # raise text positions
    #     pos[p][1] += 0.07

    nx.draw_networkx_labels(G, pos, labels)
    plt.show()


class RebuildStateIR(CfgDfs):
    def append(self, node):
        self.parent.value.stmts.append(node)

    def enter_Statement(self, node):
        self.append(node.value)

    # TODO: Why are *Sink nodes defined as statements?
    def enter_BaseBlockSink(self, node):
        pass

    def enter_HDLBlockSink(self, node):
        pass

    def enter_Module(self, block):
        block.value = type(block.value)(states=block.value.states)

    def enter_BaseBlock(self, block):
        block.value = type(block.value)()

    def exit_BaseBlock(self, block):
        if self.scopes:
            self.append(block.value)

    def enter_FuncBlock(self, block):
        irnode = block.value
        block.value = type(irnode)(irnode.args, irnode.name, irnode.ret_dtype)

    def enter_HDLBlock(self, block):
        block.value = type(block.value)()

    def exit_HDLBlock(self, block):
        self.append(block.value)

    def enter_Branch(self, block):
        block.value = type(block.value)(test=block.value.test)

    def exit_Branch(self, block):
        self.parent.value.add_branch(block.value)


class LoopBreaker(CfgDfs):
    def __init__(self, ctx, loops):
        self.ctx = ctx
        self.loops = loops
        super().__init__()

    def LoopBlock(self, node):
        skip = self.enter(node)
        if not skip:
            self.scopes.append(node)
            self.visit(node.next[0])
            self.scopes.pop()

        self.exit(node)

        return self.visit(node.next[0])

    # TODO: How to detect registers used uninitialized?
    def enter_LoopBlock(self, node):
        test = node.value.test
        loop_block = node.value
        node.value = ir.LoopBody(state_id=len(self.loops) + 1)
        jump = Node(ir.Jump('state', len(self.loops) + 1))

        self.loops.append(node)
        insert_node_before(node.sink, jump)
        cond_wrap(jump, jump, test)

        sink_id = id(node.sink.value)
        self.ctx.reaching[id(node.value)] = self.ctx.reaching.get(id(loop_block), None)
        self.ctx.reaching[id(jump.value)] = self.ctx.reaching.get(sink_id, None)
        self.ctx.reaching[id(jump.prev[0].value)] = self.ctx.reaching.get(sink_id, None)

        node.sink.value = ir.BaseBlockSink()

        loop_exit = node.next[1]
        loop_entry = node.prev[0]

        node.prev = [loop_entry]
        loop_entry.next = [node]
        node.next = [node.next[0]]

        node.sink.next = [loop_exit]
        loop_exit.prev = [node.sink]


class LoopLocality(CfgDfs):
    def __init__(self, ctx, loop, reaching_nodes):
        self.ctx = ctx
        self.loop = loop
        self.non_local = False
        self.reaching_nodes = set()
        self.loop_setup_vars = set()
        for n in reaching_nodes[self.loop]['in']:
            v = VariableFinder()
            v.visit(n.value)

            self.loop_setup_vars |= v.variables

        self.within = False
        super().__init__()

    def enter_LoopBody(self, node):
        if node is self.loop:
            self.within = True

    def exit_LoopBody(self, node):
        if node is self.loop:
            # TODO: We could exit immediatelly with exception
            self.within = False

    def enter_AssignValue(self, node):
        if self.within:
            # Node within the Loop influences statements before the loop entry.
            # This loop cannot be easily overlapped with base state
            if any(name in self.loop_setup_vars
                   for name, _ in self.ctx.reaching[id(node.value)]['gen']):
                # TODO: We could exit immediatelly with exception
                self.non_local = True


class StateIsolator(CfgDfs):
    def __init__(self, ctx, entry=None, exits=None, state_num=None):
        super().__init__()
        self.entry = entry
        if exits is None:
            exits = {}
        self.state_num = state_num
        self.exits = exits
        self.ctx = ctx
        self.node_map = {}
        self.entry_scope = None
        self.isolated = None
        self.reaching = self.ctx.reaching
        self.cpmap = {}

    def copy(self, node):
        source = self.node_map.get(node.source, None)

        cp_val = copy(node.value)

        cp_node = Node(cp_val, source=source)
        self.cpmap[node] = cp_node

        self.node_map[node] = cp_node
        cp_node.prev = [self.node_map[p] for p in node.prev if p in self.node_map]

        for n in cp_node.prev:
            n.next.append(cp_node)

        # TODO: find different way to do this, too coupled
        if id(node.value) in self.reaching:
            self.reaching[id(cp_val)] = self.reaching[id(node.value)]
        else:
            for p in cp_node.prev:
                if id(p.value) in self.reaching:
                    self.reaching[id(cp_val)] = self.reaching[id(p.value)]

        return cp_node

    def enter_BaseBlock(self, node):
        self.enter_Statement(node)

    def enter_ModuleSink(self, node):
        sink = Node(ir.ModuleSink(), prev=[node.prev[0]])
        sink = self.copy(sink)
        sink.source = self.isolated
        sink.source.sink = sink

    def enter_HDLBlockSink(self, node):
        if node.source in self.cpmap:
            self.copy(node)
        else:
            for p in node.prev:
                if p in self.node_map:
                    self.node_map[node] = self.node_map[p]

    def enter_BranchSink(self, node):
        if node.source in self.cpmap:
            self.copy(node)
        else:
            for p in node.prev:
                if p in self.node_map:
                    self.node_map[node] = self.node_map[p]

    def enter_Module(self, node):
        pass

    def enter_Statement(self, node):
        if not self.cpmap:
            if self.entry is not None:
                if node.value is not self.entry and node is not self.entry:
                    return

            self.isolated = Node(ir.Module())
            # TODO: Find better way to detect when ir.Module is entry point for isolation
            if self.entry is None and isinstance(node.prev[0].value, ir.Module):
                # node.prev[0] is ir.Module
                self.isolated.value.states = node.prev[0].value.states[:]

            self.entry = node

        if node in self.exits:
            prev = node.prev[0]
            cp_prev = self.node_map[prev]
            cp_prev.next.clear()

            test = self.exits[node]
            # exit_jump = Node(ir.AssignValue(self.ctx.ref('_state'), ir.ResExpr(self.state_num)))
            # break_stmt = Node(ir.Await(ir.res_false), prev=[exit_jump])
            return_stmt = Node(ir.Jump('state', self.state_num))
            self.state_num += 1

            if test == ir.res_true:
                return_stmt.prev = [prev]
                return_stmt = self.copy(return_stmt)
                self.node_map[self.parent.sink.prev[0]] = return_stmt
                return True

            cond_blk = Node(ir.HDLBlock(), prev=[prev])
            if_branch = Node(ir.Branch(test=test), prev=[cond_blk])
            return_stmt.prev = [if_branch]
            if_sink = Node(ir.BranchSink(), source=if_branch, prev=[return_stmt])

            else_branch = Node(ir.Branch(), prev=[cond_blk])
            else_sink = Node(ir.BranchSink(), source=else_branch, prev=[else_branch])
            cond_sink = Node(ir.HDLBlockSink(), source=cond_blk, prev=[if_sink, else_sink])

            cond_blk = self.copy(cond_blk)

            if_branch = self.copy(if_branch)
            return_stmt = self.copy(return_stmt)
            if_sink = self.copy(if_sink)

            else_branch = self.copy(else_branch)
            else_sink = self.copy(else_sink)

            cond_sink = self.copy(cond_sink)

            self.node_map[node] = cond_sink
            return

        self.copy(node)

        if self.entry is node:
            # When we enter in the middle of the graph for isolation, link to
            # previous node of entry point needs to point to new Module() node
            self.node_map[node].prev = [self.isolated]
            self.isolated.next = [self.node_map[node]]



def isolate(ctx, entry, exits=None, state_num=None):
    v = StateIsolator(ctx, exits=exits, state_num=state_num)
    v.visit(entry)
    return v.isolated


def find_statement_node(cfg, stmt):
    """Given a CFG with outgoing links, create incoming links."""
    seen = set()
    to_see = [cfg]
    while to_see:
        node = to_see.pop()

        if node.value is stmt:
            return node

        seen.add(node)
        for succ in node.next:
            if succ not in seen:
                to_see.append(succ)


def append_state_epilog(cfg, ctx):
    insert_node_before(cfg.sink, Node(ir.Jump('state', 0)))


def prepend_state_prolog(cfg, ctx, in_scope):
    source = cfg
    for name, i in ctx.intfs.items():
        if f'{name}.data' not in in_scope:
            continue

        if i.dtype.direction == 1 and in_scope.get(f'{name}.ready', False):
            continue

        hold = Node(ir.AssignValue(ir.Component(ctx.ref(name), 'data'), in_scope[f'{name}.data']),
                    prev=[source])
        hold.next = source.next

        source.next = [hold]
        hold.next[0].prev = [hold]

        source = hold


def print_cfg_ir(cfg):
    v = RebuildStateIR()
    v.visit(cfg)
    print(cfg.value)


from ..cfg import Forward


class ReachingNodes(Forward):
    """Perform reaching definition analysis.

  Each statement is annotated with a set of (variable, definition) pairs.

  """
    def __init__(self):
        def definition(node, incoming):
            kill = frozenset()
            gen = frozenset((node, node))

            # if isinstance(node.value, ir.BaseBlockSink) and isinstance(
            #         node.source.value, ir.LoopBody):
            #     gen = frozenset((node.source.value.state_id, node.source.value))
            # else:
            #     gen = frozenset()

            return gen, kill

        super(ReachingNodes, self).__init__('definitions', definition)


class Piggyback(CfgDfs):
    def __init__(self, ctx, state_id):
        self.state_id = state_id
        self.ctx = ctx
        super().__init__()

    def within_state(self):
        for s in reversed(self.scopes):
            if isinstance(s.value, ir.LoopBody):
                if s.value.state_id == self.state_id:
                    return True
        else:
            return False

    def enter_RegReset(self, node):
        if self.within_state():
            self.ctx.reset_states[node.value.target.name].add(self.state_id)


class LoopBodyFinder(CfgDfs):
    def __init__(self):
        super().__init__()
        self.loops = {}

    def enter_LoopBody(self, node):
        self.loops[node.value.state_id] = node

def discover_piggieback_states(cfg, ctx, scope_map, state_id, reaching_nodes, state_cfg):
    # cfg_temp = isolate(ctx, cfg)

    v_iso = StateIsolator(ctx)
    v_iso.visit(cfg)
    cfg_temp = v_iso.isolated

    v = ReachingNodes()
    v.visit(cfg_temp)
    reaching_nodes = v.reaching

    v = ExploreState(ctx, scope_map, state_id)
    v.visit(cfg_temp)

    if not v.states:
        return [], []

    if all(s in state_cfg for s in v.states):
        return [], []

    piggy_states = v.states

    v = cfgutil.ReachingDefinitions()
    v.visit(cfg)
    reaching_loops = v.reaching

    v = LoopBodyFinder()
    v.visit(cfg)
    loops = v.loops

    non_local = []
    piggied = []
    for child_state in piggy_states:
        loop_cfg = loops[child_state]

        v = LoopLocality(ctx, v_iso.node_map[loop_cfg], reaching_nodes)
        v.visit(cfg_temp)

        # TODO: Optimization currently disabled for loops with multiple states
        if v.non_local:
            non_local.append(child_state)
            continue

        # TODO: Maybe it should go up the "kill" tree and do this for all nodes?
        for name, in_node in reaching_loops[loop_cfg]['in']:
            if name in ctx.scope and isinstance(ctx.scope[name],
                                                ir.Variable) and not ctx.scope[name].reg:
                continue

            if isinstance(in_node.value, ir.RegReset):
                continue

            # If the register is updated withing the loop
            for out_name, out_node in reaching_loops[loop_cfg.sink]['in']:
                if out_name == name:
                    break

            if in_node is not out_node:
                state_test = ir.BinOpExpr((ctx.ref('_state'), ir.ResExpr(child_state)),
                                          ir.opc.NotEq)
                cond_wrap(in_node, in_node, state_test)

        # state_cfg[child_state] = cfg
        Piggyback(ctx, child_state).visit(cfg)
        piggied.append(child_state)

    return non_local, piggied


def schedule(cfg, ctx):
    ctx.scope['_state'] = ir.Variable(
        '_state',
        val=ir.ResExpr(Uint[1](0)),
        reg=True,
    )

    loops = []
    LoopBreaker(ctx, loops).visit(cfg)

    v = ReachingNodes()
    v.visit(cfg)
    reaching_nodes = v.reaching

    state_cfg = {0: cfg}

    isolated_loops = {}
    for l in loops:
        isolated_loops[l.value.state_id] = isolate(ctx, l)

    state_in_scope = [{} for _ in range(1 + len(loops))]
    piggied = set()
    i = 0
    order = list(state_cfg.keys())
    while i < len(order):
        state_id = order[i]
        cfg = state_cfg[state_id]
        # print(f'[{state_id}]: Scoping')
        new_states = {}
        # print_cfg_ir(cfg)

        if loops:
            non_local, piggied_cur = discover_piggieback_states(cfg, ctx, state_in_scope[state_id],
                                                                state_id, reaching_nodes,
                                                                state_cfg)

        VarScope(ctx, state_in_scope, state_id, new_states).visit(cfg)
        cfg.value.states.append(state_id)

        if loops:
            for nl in non_local:
                state_cfg[nl] = isolated_loops[nl]
                order.insert(i + 1, nl)

            cfg.value.states.extend(piggied_cur)

            piggied.update(piggied_cur)

        state_num = len(state_in_scope) - len(new_states)
        if new_states:
            # print(f'[{state_id}]: Isolating')
            state_cfg[state_id] = isolate(ctx, cfg, exits=new_states, state_num=state_num)

            # draw_scheduled_cfg(state_cfg[state_id], simple=False)
            # print_cfg_ir(state_cfg[state_id])
            for si, ns in enumerate(new_states):
                new_state_id = state_num + si

                # order.insert(i + 1, len(state_cfg))
                # print(f'[{len(state_cfg)}]: Isolating')
                # state_cfg.append(isolate(ctx, ns))

                order.insert(i + 1, new_state_id)
                # print(f'[{len(state_cfg)}]: Isolating')
                state_cfg[new_state_id] = isolate(ctx, ns)
                # draw_scheduled_cfg(state_cfg[new_state_id], simple=False)

        i += 1
        if i == len(order):
            missing = (piggied | state_cfg.keys()) ^ set(range(len(state_in_scope)))
            for s in missing:
                # TODO: Can this really be done after all stmts have been inlined?
                state_cfg[s] = isolated_loops[s]
                order.append(s)

    for i, s in state_cfg.items():
        in_scope = state_in_scope[i]

        # draw_scheduled_cfg(s, simple=True)
        # draw_scheduled_cfg(s, simple=False)
        append_state_epilog(s, ctx)
        prepend_state_prolog(s, ctx, in_scope)
        ResolveBlocking(ctx).visit(s)
        # print_cfg_ir(s)
        RebuildStateIR().visit(s)

        # print(f'------------- State {i} --------------')
        # print_cfg_ir(s)

    states = {i: s.value for i, s in state_cfg.items()}

    for i, s in states.items():
        test = ir.res_false
        for j in s.states:
            state_test = ir.BinOpExpr((ctx.ref('_state'), ir.ResExpr(j)), ir.opc.Eq)
            test = ir.BinOpExpr([test, state_test], ir.opc.Or)

    # state_num = len(states)
    state_num = len(state_in_scope)
    ctx.scope['_state'].val = ir.ResExpr(Uint[bitw(state_num - 1)](0))
    ctx.scope['_state'].dtype = Uint[bitw(state_num - 1)]

    stateblock = ir.HDLBlock()
    for i, s in states.items():
        test = ir.res_false
        for j in s.states:
            state_test = ir.BinOpExpr((ctx.ref('_state'), ir.ResExpr(j)), ir.opc.Eq)
            test = ir.BinOpExpr([test, state_test], ir.opc.Or)

        stateblock.add_branch(ir.Branch(stmts=s.stmts, test=test))

    modblock = ir.Module(stmts=[stateblock])

    return modblock



import asyncio

from pygears import gear, module, GearDone, reg
from pygears.util.find import find
from pygears.sim import delta, clk


def decouple_din_setup(module):
    module.queue = asyncio.Queue(maxsize=module.params['depth'])
    if module.params['init'] is not None:
        module.queue.put_nowait(module.params['init'])


@gear(sim_setup=decouple_din_setup)
async def decouple_din(din, *, depth, init) -> None:
    try:
        async with din as d:
            await module().queue.put(d)
            while (module().queue.full()):
                await delta()
    except GearDone:
        if module().queue.empty():
            await module().queue.put(None)

        raise GearDone


def decouple_dout_setup(module):
    module.decouple_din = find('../decouple_din')


# TODO: Implement latency == 2
@gear(sim_setup=decouple_dout_setup)
async def decouple_dout(*, t, depth, latency) -> b't':

    din = module().decouple_din
    sim_map = reg['sim/map']
    sim_din = sim_map[din]

    while True:
        if din.queue.empty() and (din not in sim_map or sim_din.done):
            raise GearDone

        queue = din.queue
        data = await queue.get()

        yield data

        queue.task_done()
        await clk()


def check_depth_pow2(depth):
    import math
    # Used to specify infinite depth size
    if depth == 0:
        return True

    from pygears.typing import TypeMatchError
    if int(math.log(depth, 2)) != math.log(depth, 2):
        raise TypeMatchError(f"Decoupler depth needs to be a power of 2, got '{depth}'")

    return True


@gear(hdl={'hierarchical': False}, enablement=b'check_depth_pow2(depth)')
def decouple(din, *, depth=2, init=None, latency=1) -> b'din':
    din | decouple_din(depth=depth, init=init)
    return decouple_dout(t=din.dtype, depth=depth, latency=latency)


buff = decouple(depth=1)



import os

from pygears.hdl.templenv import TemplateEnv
from pygears.core.hier_node import HierYielderBase
from pygears.conf import reg
from pygears.util.fileio import save_file

from .util import vgen_intf, vgen_signal


class VTemplateEnv(TemplateEnv):
    def __init__(self):
        super().__init__(basedir=os.path.dirname(__file__))

        self.jenv.globals.update(vgen_intf=vgen_intf,
                                 vgen_signal=vgen_signal)

        self.snippets = self.load(self.basedir, 'snippet.j2').module


class VGenGenerateVisitor(HierYielderBase):
    def __init__(self, top, wrapper=False):
        self.vgen_map = reg['vgen/map']
        self.wrapper = wrapper
        self.top = top
        self.template_env = VTemplateEnv()

    def RTLNode(self, node):
        vgen = self.vgen_map.get(node, None)
        if vgen is not None:
            contents = vgen.get_module(self.template_env)
            yield vgen.file_basename, contents

            # wrappers not needed for verilog, hence no else
            if (self.wrapper) and (node is self.top):
                yield f'wrap_{vgen.file_name}', vgen.get_synth_wrap(
                    self.template_env)


def vgen_generate(top, conf):
    if not conf['generate']:
        return top

    v = VGenGenerateVisitor(top, conf.get('wrapper', False))
    for file_names, contents in v.visit(top):
        if contents:
            if isinstance(contents, (tuple, list)):
                for fn, c in zip(file_names, contents):
                    save_file(fn, conf['outdir'], c)
            else:
                save_file(file_names, conf['outdir'], contents)

    return top



import inspect
from pygears.conf import PluginBase, reg
from .intf import Intf
from .gear import Gear, create_hier
from .port import InPort, OutPort, HDLConsumer, HDLProducer
from .graph import has_async_producer
from copy import deepcopy, copy
from pygears.typing import is_type
from pygears.typing.base import hashabledict


def field_names_eq(a, b):
    if not hasattr(a, 'fields'):
        return True

    if a.fields != b.fields:
        return False

    for aa, ab in zip(a.args, b.args):
        if not is_type(aa):
            continue

        if not field_names_eq(aa, ab):
            return False

    return True


def copy_gear_full(g: Gear, name=None):
    params = {}
    for n, val in g.params.items():
        if isinstance(val, (list, dict)):
            params[n] = val.copy()
        else:
            params[n] = val

    params['memoized'] = g

    if name:
        params['name'] = name

    cp_g = Gear(g.func, params)

    # Important! Have to keep the copy in local array, or garbage collector
    # collects unconnected port consumers
    local_in_ports = []
    for p in g.in_ports:
        cp_p = InPort(cp_g, p.index, p.basename, dtype=p.dtype)
        cp_g.in_ports.append(cp_p)
        i = Intf(p.dtype)
        local_in_ports.append(i)
        i.source(cp_p)

    # Important! Have to keep the copy in local array, or garbage collector
    # collects unconnected port consumers
    outside_intfs = []
    for p in g.out_ports:
        cp_p = OutPort(cp_g, p.index, p.basename, dtype=p.dtype)
        cp_g.out_ports.append(cp_p)

        if not g.hierarchical:
            inside_intf = Intf(p.dtype)
            if hasattr(p.producer, 'var_name'):
                inside_intf.var_name = p.producer.var_name

            inside_intf.connect(cp_p)

            if has_async_producer(p):
                inside_intf.source(HDLProducer())

        i = Intf(p.dtype)
        outside_intfs.append(i)
        i.source(cp_p)

    cp_map = {g: cp_g}
    cp_outside_intf_map = {}
    with create_hier(cp_g):
        for c in g.child:
            cp_map[c], cp_outside_intf_map[c] = copy_gear_full(c)

    def copy_port_connection(p):
        cp_prod_gear = cp_map[p.gear]
        cp_prod = getattr(cp_prod_gear, f'{p.direction}_ports')[p.index]
        cp_intf = cp_prod.consumer

        intf = p.consumer
        if intf is None:
            return

        for cons in intf.consumers:
            if isinstance(cons, HDLConsumer):
                cp_intf.connect(HDLConsumer())
                continue

            cp_cons_gear = cp_map[cons.gear]
            cp_cons = getattr(cp_cons_gear, f'{cons.direction}_ports')[cons.index]

            cp_intf.connect(cp_cons)

    for p in g.in_ports:
        copy_port_connection(p)

    for c in g.child:
        for p in c.out_ports:
            copy_port_connection(p)

    return cp_g, outside_intfs


# TODO: Handle unpack gears that create ccat in parents space
def copy_gear(mem_gear: Gear, args, kwds, name, intf_mapping, kwd_intfs):
    gear_inst, outside_intfs = copy_gear_full(mem_gear, name)
    in_num = len(gear_inst.in_ports)

    for pi, ii in intf_mapping.items():
        if pi < len(gear_inst.in_ports):
            kwd_intfs[ii].connect(gear_inst.in_ports[pi])
        else:
            kwd_intfs[ii].source(gear_inst.out_ports[pi - in_num])

    for key in kwds:
        if key in reg['gear/params/extra']:
            gear_inst.params[key] = kwds[key]

    for name, val in mem_gear.const_args.items():
        from pygears.lib import const
        const(val=val, intfs=[args[name]])

    for i, intf in enumerate(args.values()):
        p = gear_inst.in_ports[i]
        assert p.producer is None
        intf.connect(p)

    out_intfs = tuple(o for i, o in enumerate(gear_inst.outputs)
                      if (i + in_num) not in intf_mapping)

    return gear_inst, out_intfs


class ContainerVisitor:
    def __init__(self, kwd_intfs):
        self.kwd_intfs = kwd_intfs
        self.cur_addr = []

    def visit(self, obj):
        for base_class in inspect.getmro(obj.__class__):
            if hasattr(self, f'visit_{base_class.__name__}'):
                return getattr(self, f'visit_{base_class.__name__}')(obj)
        else:
            return self.generic_visit(obj)

    def visit_tuple(self, obj):
        hsh = []
        for i, o in enumerate(obj):
            hsh.append((self.visit(o), i))

        return hash(tuple(hsh))

    def visit_NoneType(self, obj):
        return hash('None')

    visit_list = visit_tuple

    def visit_dict(self, obj):
        hsh = []
        for i, o in obj.items():
            hsh.append((self.visit(o), i))

        return hash(tuple(hsh))

    def visit_Intf(self, obj):
        self.kwd_intfs.append(obj)
        return hash(obj.dtype)

    def visit_slice(self, obj):
        return hash(obj.__reduce__())

    def generic_visit(self, obj):
        return hash(obj)


def make_gear_call_hash(func, args, const_args, kwds, fix_intfs):
    user_kwds = kwds.copy()
    for key in reg['gear/params/extra']:
        if key != '__hdl__':
            user_kwds.pop(key, None)

    try:
        kwd_intfs = []
        v = ContainerVisitor(kwd_intfs)
        kwd_hsh = v.visit(user_kwds)

        v = ContainerVisitor(kwd_intfs)
        fixi_hsh = v.visit(fix_intfs)

        # Cannot use hash(func) since it is dependent on the id(func)
        # TODO: This func hashing will not work for lambda
        hsh = kwd_hsh ^ fixi_hsh ^ hash((
            func.__code__.co_name,
            func.__code__.co_firstlineno,
            func.__code__.co_filename,
            tuple(a.dtype if isinstance(a, Intf) else a for a in args.values()),
            hashabledict(const_args),
        ))

        return hsh, tuple(kwd_intfs)

    except TypeError:
        return None, None

def gear_inst_hash(g):

    kwds = g.params.copy()
    for p in g.in_ports + g.out_ports:
        if p.basename in kwds:
            del kwds[p.basename]

    return make_gear_call_hash(g.func, g.args, g.const_args, kwds, ())[0]

def get_memoized_gear(func, args, const_args, kwds, fix_intfs, name):
    key, kwd_intfs = make_gear_call_hash(func, args, const_args, kwds, fix_intfs)

    if key is None:
        return None, None, None

    gear_memoize = reg['gear/memoize_map']
    if key not in gear_memoize:
        return None, None, (key, kwd_intfs)

    if name == 'apply':
        return None, None, None

    mem_gear, intf_mapping = gear_memoize[key]

    gear_inst, outputs = copy_gear(mem_gear, args, kwds, name, intf_mapping, kwd_intfs)
    return gear_inst, outputs, None


def memoize_gear(gear_inst, key):
    if key is None:
        return

    key, kwd_intfs = key

    intf_mapping = {}
    for i, intf in enumerate(kwd_intfs):
        for p in gear_inst.in_ports:
            if intf is p.producer:
                intf_mapping[p.index] = i

        for p in gear_inst.out_ports:
            if intf is p.consumer:
                intf_mapping[p.index + len(gear_inst.in_ports)] = i

    gear_memoize = reg['gear/memoize_map']
    gear_memoize[key] = (gear_inst, intf_mapping)


class GearMemoizePlugin(PluginBase):
    @classmethod
    def bind(cls):
        reg['gear/memoize_map'] = {}

    @classmethod
    def reset(cls):
        reg['gear/memoize_map'] = {}
